# -*- coding: utf-8 -*-

import binarybrain      as bb
import binarybrain.core as core
import numpy as np
from typing import List

class Tensor(bb.Object):
    """Tensor class

        多次元データ構造。

    Args:
        shape (list[int]):  Shape of created array
        dtype (int): Data type
        host_only (bool): flag of host only
    """
    
    def __init__(self, shape: List[int]=None, *, dtype = bb.DType.FP32, host_only=False, core_tensor=None):
        if core_tensor is None:
            if shape is not None:
                core_tensor = core.Tensor(shape, dtype.value, host_only)
        super(Tensor, self).__init__(core_object=core_tensor)

    @staticmethod
    def from_core(core_tensor):
        new_tensor = Tensor(shape=None, core_tensor=core_tensor)
        return new_tensor
    
    def is_host_only(self) -> bool:
        return self.get_core().is_host_only()

    def get_type(self) -> int:
        """データ型取得
        
        Returns:
            data type.
        """
        return bb.DType(self.get_core().get_type())
    
    def get_shape(self) -> List[int]:
        """データのシェイプ取得
        
        Returns:
            shape
        """
        return self.get_core().get_shape()
    
    def numpy(self) -> np.ndarray:
        """NumPy の ndarray に変換
        
        Returns:
            ndarray (array)
        """

        dtype = self.get_core().get_type()
        if dtype == bb.DType.FP32:
            return self.get_core().numpy_fp32()
        elif dtype == bb.DType.FP64:
            return self.get_core().numpy_fp64()
        elif dtype == bb.DType.INT8:
            return self.get_core().numpy_int8()
        elif dtype == bb.DType.INT16:
            return self.get_core().numpy_int16()
        elif dtype == bb.DType.INT32:
            return self.get_core().numpy_int32()
        elif dtype == bb.DType.INT64:
            return self.get_core().numpy_int64()
        elif dtype == bb.DType.UINT8:
            return self.get_core().numpy_uint8()
        elif dtype == bb.DType.UINT16:
            return self.get_core().numpy_uint16()
        elif dtype == bb.DType.UINT32:
            return self.get_core().numpy_uint32()
        elif dtype == bb.DType.UINT64:
            return self.get_core().numpy_uint64()

    def set_numpy(self, ndarray: np.ndarray):
        dtype = self.get_core().get_type()
        assert(bb.dtype_numpy_to_bb(ndarray.dtype) == dtype)
        assert(ndarray.shspe == self.get_shape)
        # 未実装

    @staticmethod
    def from_numpy(ndarray: np.ndarray, host_only=False):
        """NumPy から生成
        
        Args:
            ndarray (ndarray): array of NumPy
            host_only (bool): flag of host only
        """
        
        if not ndarray.flags['C_CONTIGUOUS']:
            ndarray = ndarray.copy(order='C')
        if ndarray.dtype == np.float32:
            core_tensor = bb.core.Tensor.from_numpy_fp32(ndarray, host_only)
        elif ndarray.dtype == np.float64:
            core_tensor = bb.core.Tensor.from_numpy_fp64(ndarray, host_only)
        elif ndarray.dtype == np.int8:
            core_tensor = bb.core.Tensor.from_numpy_int8(ndarray, host_only)
        elif ndarray.dtype == np.int16:
            core_tensor = bb.core.Tensor.from_numpy_int16(ndarray, host_only)
        elif ndarray.dtype == np.int32:
            core_tensor = bb.core.Tensor.from_numpy_int32(ndarray, host_only)
        elif ndarray.dtype == np.int64:
            core_tensor = bb.core.Tensor.from_numpy_int64(ndarray, host_only)
        elif ndarray.dtype == np.uint8:
            core_tensor = bb.core.Tensor.from_numpy_uint8(ndarray, host_only)
        elif ndarray.dtype == np.uint16:
            core_tensor = bb.core.Tensor.from_numpy_uint16(ndarray, host_only)
        elif ndarray.dtype == np.uint32:
            core_tensor = bb.core.Tensor.from_numpy_uint32(ndarray, host_only)
        elif ndarray.dtype == np.uint64:
            core_tensor = bb.core.Tensor.from_numpy_uint64(ndarray, host_only)
        else:
            core_tensor = None
            raise TypeError("unsupported")
        
        return Tensor(core_tensor=core_tensor)


    def __add__(self, x):
        if type(x) == Tensor:
            core_tensor = self.get_core() + x.get_core()
        else:
            core_tensor = self.get_core() + float(x)
        return Tensor(core_tensor=core_tensor)

    def __sub__(self, x):
        if type(x) == Tensor:
            core_tensor = self.get_core() - x.get_core()
        else:
            core_tensor = self.get_core() - float(x)
        return Tensor(core_tensor=core_tensor)

    def __mul__(self, x):
        if type(x) == Tensor:
            core_tensor = self.get_core() * x.get_core()
        else:
            core_tensor = self.get_core() * float(x)
        return Tensor(core_tensor=core_tensor)

    def __truediv__(self, x):
        if type(x) == Tensor:
            core_tensor = self.get_core() / x.get_core()
        else:
            core_tensor = self.get_core() / float(x)
        return Tensor(core_tensor=core_tensor)

    def __radd__(self, x):
        if type(x) == Tensor:
            core_tensor = x.get_core() + self.get_core() 
        else:
            core_tensor = float(x) + self.get_core()
        return Tensor(core_tensor=core_tensor)

    def __rsub__(self, x):
        if type(x) == Tensor:
            core_tensor = x.get_core() - self.get_core() 
        else:
            core_tensor = float(x) - self.get_core()
        return Tensor(core_tensor=core_tensor)

    def __rmul__(self, x):
        if type(x) == Tensor:
            core_tensor = x.get_core() * self.get_core() 
        else:
            core_tensor = float(x) * self.get_core()
        return Tensor(core_tensor=core_tensor)

    def __rtruediv__(self, x):
        if type(x) == Tensor:
            core_tensor = x.get_core() / self.get_core() 
        else:
            core_tensor = float(x) / self.get_core()
        return Tensor(core_tensor=core_tensor)
    
    def __iadd__(self, x):
        core_tensor = self.get_core()
        if type(x) == Tensor:
            core_tensor += x.get_core()
        else:
            core_tensor += float(x)
        return self
    
    def __isub__(self, x):
        core_tensor = self.get_core()
        if type(x) == Tensor:
            core_tensor -= x.get_core() 
        else:
            core_tensor -= float(x)
        return self
    
    def __imul__(self, x):
        core_tensor = self.get_core()
        if type(x) == Tensor:
            core_tensor *= x.get_core() 
        else:
            core_tensor *= float(x)
        return self
    
    def __itruediv__(self, x):
        core_tensor = self.get_core()
        if type(x) == Tensor:
            core_tensor /= x.get_core() 
        else:
            core_tensor /= float(x)
        return self
    


# -*- coding: utf-8 -*-

import binarybrain      as bb
import binarybrain.core as core
import numpy as np
from typing import List


class Variables():
    """Variables class

       学習の為の Optimizer と実際の学習ターゲットの変数の橋渡しに利用されるクラス。
       内部的には各モデル内の重みや勾配を保有する Tensor をまとめて保持している。
    """
    
    def __init__(self):
        self.variables = core.Variables()

    @staticmethod
    def from_core(variables):
        new_variables = Variables()
        new_variables.variables = variables
        return new_variables
    
    def get_core(self):
        return self.variables
    
    def append(self, variables):
        """ 変数を追加

        Args:
            variables (Variables) : 追加する変数
        """
        self.variables.push_back(variables.get_core())

    def get_size(self):
        return self.variables.get_size()
    
    def at(self, item):
        return self.variables.at(item)
    
    def __len__(self):
        return self.variables.get_size()
        
    def __getitem__(self, item):
        return self.variables.at(item)
    



# -*- coding: utf-8 -*-

import numpy as np

import binarybrain      as bb
import binarybrain.core as core


# ----- LUT Layer -----

def make_verilog_lut_layers(module_name: str, net, device=""):
    layers = bb.get_model_list(net, flatten=True)
    core_layers = []
    for layer in layers:
        core_layers.append(layer.get_core())
    return core.make_verilog_lut_layers(module_name, core_layers, device)


def dump_verilog_lut_layers(f, module_name: str, net, device=""):
    ''' make verilog source of LUT layers
        変換できないモデルは影響ない層とみなして無視するので注意
    
        Args:
            f (StreamIO) : 出力先ストリーム
            module_name (str): モジュール名
            net (Model): 変換するネット
            
        Returns:
            Verilog source code (str)
    '''
    f.write(make_verilog_lut_layers(module_name=module_name, net=net, device=device))

def export_verilog_lut_layers(file_name: str, module_name: str, net):
    with open(file_name, 'w') as f:
        dump_verilog_lut_layers(f, module_name, net)



# ----- Convolutional LUT Layer -----

def make_verilog_lut_cnv_layers(module_name: str, net, device=""):
    layers = bb.get_model_list_for_rtl(net)
    core_layers = []
    for layer in layers:
        core_layers.append(layer.get_core())
    return core.make_verilog_lut_cnv_layers(module_name, core_layers, device)


def dump_verilog_lut_cnv_layers(f, module_name: str, net, device=""):
    ''' dump verilog source of Convolutional LUT layers
        
        畳み込み層を含むネットを AXI4 Stream Video 形式のVerilogソースコードして
        出力する。
        縮小を伴う MaxPooling 層は最後に1個だけ挿入を許される

        Args:
            f (StreamIO) : 出力先ストリーム
            module_name (str): モジュール名
            net (Model): 変換するネット
    '''
    f.write(make_verilog_lut_cnv_layers(module_name, net, device))

def export_verilog_lut_cnv_layers(file_name: str, module_name: str, net, device=""):
    with open(file_name, 'w') as f:
        dump_verilog_lut_cnv_layers(f, module_name, net, device)



# ----- For Simuration -----

def __dump_bin_digit(f, v):
    if v:
        f.write('1')
    else:
        f.write('0')

def __dump_bin_int(f, v, digits):
    for i in range(digits):
        __dump_bin_digit(f, ((v >> (digits-1-i)) & 1))

def __dump_bin_img(f, img):
    img = np.array(img).flatten()[::-1]
    for v in img:
        __dump_bin_digit(f, v > 0.5)

def dump_verilog_readmemb_image_classification(f, loader, *, class_digits=8):
    """verilog用データダンプ
    verilog の $readmemb() での読み込み用データ作成

    クラスID + 画像データの形式で出力する

    Args:
        f (StreamIO): 出力先
        loader (Loader): モジュール名
        class_digits (int): クラス分類のbit数
    """

    for images, labels in loader:
        for x, t in zip(images, labels):
            __dump_bin_int(f, t, class_digits)
            f.write('_')
            __dump_bin_img(f, x)
            f.write('\n')



def make_image_tile(rows, cols, img_gen):
    """画像をタイル状に並べて大きな画像にする

       学習用の c, h, w 順の画像データをタイル状に結合する

    Args:
        rows (int)): 縦の結合枚数
        cols (int)): 横の結合枚数
        gen (ndarray): 画像を返すジェネレータ

    Returns:
        img (ndarray) : 作成した画像
    """

    def make_image_tile_h(cols, img_gen):
        img = img_gen.__next__()
        for _ in range(1, cols):
            img = np.concatenate((img, img_gen.__next__()), axis=img.ndim-1)
        return img

    img = make_image_tile_h(cols, img_gen)
    for _ in range(1, rows):
        img = np.concatenate((img, make_image_tile_h(cols, img_gen)), axis=img.ndim-2)
    return img


def write_ppm(fname, img):
    """ppmファイルの出力

        学習用の c, h, w 順の画像データを ppm形式で保存する

    Args:
        fname (str): 出力ファイル名
        img (ndarray): モジュール名
    """

    # gray to color
    if img.ndim == 3 and img.shape[0] == 1:
        img = np.tile(img, (3, 1, 1))
    elif img.ndim == 2:
        img = np.stack((img, img, img))

    # shpe check    
    assert(img.ndim == 3 and img.shape[0] >= 3)
    
    # write
    with open(fname, 'w') as f:
        f.write('P3\n')
        f.write('%d %d\n' % (img.shape[2], img.shape[1]))
        f.write('255\n')
        img = img.transpose(1, 2, 0).reshape(-1, 3)
        for v in img:
            f.write('%d %d %d\n' % (v[0], v[1], v[2]))




# -*- coding: utf-8 -*-

import binarybrain      as bb
import binarybrain.core as core
import numpy as np
from enum import IntEnum


class DType(IntEnum):
    """データ型定義
    """

    BIT    = (0x0000 + 1)
    BINARY = (0x0000 + 2)
    FP16   = (0x0100 + 16)
    FP32   = (0x0100 + 32)
    FP64   = (0x0100 + 64)
    INT8   = (0x0200 + 8)
    INT16  = (0x0200 + 16)
    INT32  = (0x0200 + 32)
    INT64  = (0x0200 + 64)
    UINT8  = (0x0300 + 8)
    UINT16 = (0x0300 + 16)
    UINT32 = (0x0300 + 32)
    UINT64 = (0x0300 + 64)


class Border(IntEnum):
    CONSTANT    = 0
    REFLECT     = 1
    REFLECT_101 = 2
    REPLICATE   = 3
    WRAP        = 4


def dtype_numpy_to_bb(dtype):
    if   dtype == np.float32: return core.TYPE_FP32
    elif dtype == np.float64: return core.TYPE_FP64
    elif dtype == np.int8:    return core.TYPE_INT8
    elif dtype == np.int16:   return core.TYPE_INT16
    elif dtype == np.int32:   return core.TYPE_INT32
    elif dtype == np.int64:   return core.TYPE_INT64
    elif dtype == np.uint8:   return core.TYPE_UINT8
    elif dtype == np.uint16:  return core.TYPE_UINT16
    elif dtype == np.uint32:  return core.TYPE_UINT32
    elif dtype == np.uint64:  return core.TYPE_UINT64
    return None

def dtype_bb_to_numpy(dtype):
    if   dtype == core.TYPE_FP32:   return np.float32
    elif dtype == core.TYPE_FP64:   return np.float64
    elif dtype == core.TYPE_INT8:   return np.int8
    elif dtype == core.TYPE_INT16:  return np.int16
    elif dtype == core.TYPE_INT32:  return np.int32
    elif dtype == core.TYPE_INT64:  return np.int64
    elif dtype == core.TYPE_UINT8:  return np.uint8
    elif dtype == core.TYPE_UINT16: return np.uint16
    elif dtype == core.TYPE_UINT32: return np.uint32
    elif dtype == core.TYPE_UINT64: return np.uint64
    return None

def dtype_to_name(dtype):
    if   dtype == core.TYPE_BIT:    return 'bit'
    elif dtype == core.TYPE_FP32:   return 'fp32'
    elif dtype == core.TYPE_FP64:   return 'fp64'
    elif dtype == core.TYPE_INT8:   return 'int8'
    elif dtype == core.TYPE_INT16:  return 'int16'
    elif dtype == core.TYPE_INT32:  return 'int32'
    elif dtype == core.TYPE_INT64:  return 'int64'
    elif dtype == core.TYPE_UINT8:  return 'uint8'
    elif dtype == core.TYPE_UINT16: return 'uint16'
    elif dtype == core.TYPE_UINT32: return 'uint32'
    elif dtype == core.TYPE_UINT64: return 'uint64'
    return None

def dtype_from_name(name):
    if   name == 'bit':    return core.TYPE_BIT
    elif name == 'fp32':   return core.TYPE_FP32
    elif name == 'fp64':   return core.TYPE_FP64
    elif name == 'int8':   return core.TYPE_INT8
    elif name == 'int16':  return core.TYPE_INT16
    elif name == 'int32':  return core.TYPE_INT32
    elif name == 'int64':  return core.TYPE_INT64
    elif name == 'uint8':  return core.TYPE_UINT8
    elif name == 'uint16': return core.TYPE_UINT16
    elif name == 'uint32': return core.TYPE_UINT32
    elif name == 'uint64': return core.TYPE_UINT64 
    return None





# -*- coding: utf-8 -*-

import binarybrain      as bb
import binarybrain.core as core
import numpy as np
from typing import List



class Metrics(bb.Object):
    """Metrics class
        評価関数の基底クラス
    """
    
    def __init__(self, core_metrics=None):
        super(Metrics, self).__init__(core_object=core_metrics)
        
    def clear(self):
        """値のクリア

           集計をクリアする。通常 epoch の単位でクリアして再集計を行う
        """
        self.get_core().clear()
    
    def get(self):
        """値の取得

        Returns:
            metrics(float) : 現在までの損失値を返す
        """
        return self.get_core().get_metrics()

    def calculate(self, y_buf, t_buf):
        """評価の計算

        Args:
            y_buf (FrameBuffer): forward演算結果
            t_buf (FrameBuffer): 教師データ
        """
        return self.get_core().calculate_metrics(y_buf.get_core(), t_buf.get_core())

    def get_metrics_string(self):
        """評価対象の文字列取得

        評価関数ごとに評価値の単位が異なるため計算しているものの文字列を返す
        平均二乗誤差(MSE)であったり、認識率(accuracy)であったり
        getで得られる値を、表示やログで表す際に利用できる

        Args:
            metrics_string (str): 評価対象の文字列取得
        """
        return self.get_core().get_metrics_string()


class MetricsMeanSquaredError(Metrics):
    """MetricsMeanSquaredError class

       平均二乗誤差の評価関数
       教師信号との平均二乗誤差を計算する
    """
    
    def __init__(self, dtype=bb.DType.FP32):
        core_metrics = bb.search_core_object('MetricsMeanSquaredError', [dtype]).create()
        super(MetricsMeanSquaredError, self).__init__(core_metrics=core_metrics)


class MetricsCategoricalAccuracy(Metrics):
    """MetricsCategoricalAccuracy class

       多クラス分類用の評価関数
       一致率を accuracy として計算する
    """
    
    def __init__(self, dtype=bb.DType.FP32):
        core_metrics = bb.search_core_object('MetricsCategoricalAccuracy', [dtype]).create()
        super(MetricsCategoricalAccuracy, self).__init__(core_metrics=core_metrics)


class MetricsBinaryCategoricalAccuracy(Metrics):
    """MetricsBinaryCategoricalAccuracy class

       2クラス分類用の評価関数
       一致率を accuracy として計算する
    """
    
    def __init__(self, dtype=bb.DType.FP32):
        core_metrics = bb.search_core_object('MetricsBinaryCategoricalAccuracy', [dtype]).create()
        super(MetricsBinaryCategoricalAccuracy, self).__init__(core_metrics=core_metrics)




# -*- coding: utf-8 -*-

import binarybrain      as bb
import binarybrain.core as core
import numpy as np
from typing import List

class FrameBuffer(bb.Object):
    """FrameBuffer class
       
        BinaryBrainでの学習データを格納する特別な型である
        バッチと対応する 1次元のframe項と、各レイヤーの入出力ノードに対応する
        多次元のnode項を有している
        numpy の ndarray に変換する際は axis=0 が frame 項となり、以降が
        node項となる

        Tensor と異なり、frame 項に対して reshape を行うことはできず、
        node 項に対しても transpose することはできない。

        node に関しては 2次元以上の shape も持ちうるが、実際のレイヤー間接続に
        際しては、畳み込みなどの次元に意味を持つノード以外では、ノード数さえ
        あっていれば接続できるものが殆どである(多くの処理系で必要とする
        flatten が省略できる)。

        host_only フラグを指定すると device(GPU側) が利用可能であっても
        host(CPU側) のみにメモリを確保する

    Args:
        frame_size (int): frame サイズ
        shape (list[int]):  node シェイプ
        dtype (int): Data type
        host_only (bool): flag of host only
    """
    
    def __init__(self, frame_size: int = 0, shape: List[int] = [], dtype = bb.DType.FP32, host_only: bool = False, core_buf=None):
        if core_buf is None:
            core_buf = core.FrameBuffer(frame_size, shape, dtype.value, host_only)
        super(FrameBuffer, self).__init__(core_object=core_buf)

    @staticmethod
    def from_core(core_buf):
        return FrameBuffer(core_buf=core_buf)

    def is_host_only(self) -> bool:
        return self.get_core().is_host_only()

    def get_type(self) -> int:
        """データ型取得
        
        Returns:
            dtype (DType)
        """
        return bb.DType(self.get_core().get_type())
    
    def get_frame_size(self) -> int:
        """get size of frame.
        
        Returns:
            frame size.
        """
        return self.get_core().get_frame_size()
    
    def get_frame_stride(self) -> int:
        """get stride of frame.
        
        Returns:
            frame stride.
        """
        return self.get_core().get_frame_stride()
        
    def get_node_size(self) -> int:
        """get size of node.
        
        Returns:
            node size.
        """
        return self.get_core().get_node_size()
    
    def get_node_shape(self) -> List[int]:
        """get shape of node.
        
        Returns:
            shape
        """
        return self.get_core().get_node_shape()
    
    def numpy(self) -> np.ndarray:
        """Convert to NumPy
        
        Args:
            shape (list[int]):  Shape of created array
            dtype (int): Data type
            host_only (bool): flag of host only
        """
        dtype = self.get_type()
        if dtype == bb.DType.BIT:
            ndarray = self.get_core().numpy_uint8()
        elif dtype == bb.DType.BINARY:
            ndarray = self.get_core().numpy_uint8()
        elif dtype == bb.DType.FP32:
            ndarray = self.get_core().numpy_fp32()
        elif dtype == bb.DType.FP64:
            ndarray = self.get_core().numpy_fp64()
        elif dtype == bb.DType.INT8:
            ndarray =  self.get_core().numpy_int8()
        elif dtype == bb.DType.INT16:
            ndarray =  self.get_core().numpy_int16()
        elif dtype == bb.DType.INT32:
            ndarray =  self.get_core().numpy_int32()
        elif dtype == bb.DType.INT64:
            ndarray =  self.get_core().numpy_int64()
        elif dtype == bb.DType.UINT8:
            ndarray =  self.get_core().numpy_uint8()
        elif dtype == bb.DType.UINT16:
            ndarray =  self.get_core().numpy_uint16()
        elif dtype == bb.DType.UINT32:
            ndarray =  self.get_core().numpy_uint32()
        elif dtype == bb.DType.UINT64:
            ndarray =  self.get_core().numpy_uint64()
        else:
            raise TypeError("unexpected dtype")

        tran = [ndarray.ndim-1] + list(range(0, ndarray.ndim-1))
        ndarray = ndarray.transpose(tran)
        if dtype != bb.DType.BIT:
            shape = list(ndarray.shape)
            shape[0] = self.get_core().get_frame_size()
            ndarray = np.resize(ndarray, shape)
        
        return ndarray
        
    @staticmethod
    def from_numpy(ndarray: np.ndarray, host_only=False):
        """Create from NumPy
        
        Args:
            ndarray (np.ndarray): array of NumPy
            host_only (bool): flag of host only
        """
        
        shape = list(ndarray.shape)
        assert(len(shape) >= 2)
        bb_dtype = bb.dtype_numpy_to_bb(ndarray.dtype)
        frame_size   = shape[0]
        frame_stride = core.FrameBuffer.calc_frame_stride(bb_dtype, frame_size) 
        shape[0] = frame_stride // core.dtype_get_byte_size(bb_dtype)
        ndarray = np.resize(ndarray, shape)
        
        tran = list(range(1, ndarray.ndim)) + [0]
        ndarray = ndarray.transpose(tran)
        ndarray = ndarray.copy(order='C')
        
        if ndarray.dtype == np.float32:
            core_buf = bb.core.FrameBuffer.from_numpy_fp32(ndarray, bb_dtype, frame_size, frame_stride, shape[1:], host_only)
        elif ndarray.dtype == np.float64:
            core_buf = bb.core.FrameBuffer.from_numpy_fp64(ndarray, bb_dtype, frame_size, frame_stride, shape[1:],host_only)
        elif ndarray.dtype == np.int8:
            core_buf = bb.core.FrameBuffer.from_numpy_int8(ndarray, bb_dtype, frame_size, frame_stride, shape[1:],host_only)
        elif ndarray.dtype == np.int16:
            core_buf = bb.core.FrameBuffer.from_numpy_int16(ndarray, bb_dtype, frame_size, frame_stride, shape[1:],host_only)
        elif ndarray.dtype == np.int32:
            core_buf = bb.core.FrameBuffer.from_numpy_int32(ndarray, bb_dtype, frame_size, frame_stride, shape[1:],host_only)
        elif ndarray.dtype == np.int64:
            core_buf = bb.core.FrameBuffer.from_numpy_int64(ndarray, bb_dtype, frame_size, frame_stride, shape[1:],host_only)
        elif ndarray.dtype == np.uint8:
            core_buf = bb.core.FrameBuffer.from_numpy_uint8(ndarray, bb_dtype, frame_size, frame_stride, shape[1:],host_only)
        elif ndarray.dtype == np.uint16:
            core_buf = bb.core.FrameBuffer.from_numpy_uint16(ndarray, bb_dtype, frame_size, frame_stride, shape[1:],host_only)
        elif ndarray.dtype == np.uint32:
            core_buf = bb.core.FrameBuffer.from_numpy_uint32(ndarray, bb_dtype, frame_size, frame_stride, shape[1:],host_only)
        elif ndarray.dtype == np.uint64:
            core_buf = bb.core.FrameBuffer.from_numpy_uint64(ndarray, bb_dtype, frame_size, frame_stride, shape[1:],host_only)
        else:
            core_buf = None
            raise TypeError("unsupported")
        return FrameBuffer(core_buf=core_buf)
    
    
    def __add__(self, x):
        if type(x) == FrameBuffer:
            core_buf = self.get_core() + x.get_core()
        else:
            core_buf = self.get_core() + float(x)
        return FrameBuffer(core_buf=core_buf)

    def __sub__(self, x):
        if type(x) == FrameBuffer:
            core_buf = self.get_core() - x.get_core()
        else:
            core_buf = self.get_core() - float(x)
        return FrameBuffer(core_buf=core_buf)

    def __mul__(self, x):
        if type(x) == FrameBuffer:
            core_buf = self.get_core() * x.get_core()
        else:
            core_buf = self.get_core() * float(x)
        return FrameBuffer(core_buf=core_buf)

    def __truediv__(self, x):
        if type(x) == FrameBuffer:
            core_buf = self.get_core() / x.get_core()
        else:
            core_buf = self.get_core() / float(x)
        return FrameBuffer(core_buf=core_buf)

    def __radd__(self, x):
        if type(x) == FrameBuffer:
            core_buf = x.get_core() + self.get_core() 
        else:
            core_buf = float(x) + self.get_core()
        return FrameBuffer(core_buf=core_buf)

    def __rsub__(self, x):
        if type(x) == FrameBuffer:
            core_buf = x.get_core() - self.get_core() 
        else:
            core_buf = float(x) - self.get_core()
        return FrameBuffer(core_buf=core_buf)

    def __rmul__(self, x):
        if type(x) == FrameBuffer:
            core_buf = x.get_core() * self.get_core() 
        else:
            core_buf = float(x) * self.get_core()
        return FrameBuffer(core_buf=core_buf)

    def __rtruediv__(self, x):
        if type(x) == FrameBuffer:
            core_buf = x.get_core() / self.get_core() 
        else:
            core_buf = float(x) / self.get_core()
        return FrameBuffer(core_buf=core_buf)
    
    def __iadd__(self, x):
        core_buf = self.get_core()
        if type(x) == FrameBuffer:
            core_buf += x.get_core() 
        else:
            core_buf += float(x)
        return self
    
    def __isub__(self, x):
        core_buf = self.get_core()
        if type(x) == FrameBuffer:
            core_buf -= x.get_core() 
        else:
            core_buf -= float(x)
        return self
    
    def __imul__(self, x):
        core_buf = self.get_core()
        if type(x) == FrameBuffer:
            core_buf *= x.get_core() 
        else:
            core_buf *= float(x)
        return self
    
    def __itruediv__(self, x):
        core_buf = self.get_core()
        if type(x) == FrameBuffer:
            core_buf /= x.get_core() 
        else:
            core_buf /= float(x)
        return self




# -*- coding: utf-8 -*-

import re
import pickle
import numpy as np
from typing import List

import binarybrain      as bb
import binarybrain.core as core


# ------- モデルリスト --------

#core_model_list = bb.get_core_subclass_dict(core.Model)
#core_model_dict = bb.get_core_subclass_dict(core.Model)

def search_core_model(model_name, dtypes=None):
    if dtypes is None:
        dtypes = []
    return bb.search_core_object(model_name, dtypes)

_model_creator_list = {}
def model_creator_regist(model_name, creator):
    _model_creator_list[model_name] = creator

def model_creator(data, name, dtypes):
    if name in  _model_creator_list:
        return _model_creator_list[name](data)
    return data, None

bb.object_creator_regist(model_creator)


# ------- 基本モデル --------

class Model(bb.Object):
    """Model class
       ネットワーク各層の演算モデルの基底クラス
       すべてのモデルはこのクラスを基本クラスに持つ

       BinaryBrain では、モデルを実際にインスタンスとして生成して組み合わせることで
       学習ネットワークを構成する。特にネットワーク内のインスタンス化された
       モデルをレイヤーという呼び方をする場合がある。
    """
    
    def __init__(self, *, core_model=None, input_shape=None, name=None):
        super(Model, self).__init__(core_object=core_model)

        self.input_shape = input_shape
        self.name = name
        if core_model is not None:
            if name is not None:
                self.set_name(name)
            if input_shape is not None:
                self.set_input_shape(input_shape)

    @classmethod
    def from_bytes(cls, data):
        data, core_model = bb.core_object_loads(data)
        return data, cls(core_model=core_model)

    def set_name(self, name: str):
        """インスタンス名の設定

           生成したモデルには任意の名前を付けることが可能であり、表示や
           保存時のファイル名などに利用することが可能である

        Args:
            name (str): 新しいインスタンス名
        """
        self.name = name
        core_model = self.get_core()
        if core_model is not None:
            core_model.set_name(name)

    def get_name(self):
        """インスタンス名の取得

           インスタンス名を取得する。名称が設定されていない場合はクラス名が返される

        Returns:
            name (str)
        """
        core_model = self.get_core()
        if core_model is not None:
            return core_model.get_name()
        else:
            if self.name:
                return self.name
            else:
                return self.get_model_name()

    def is_named(self):
        """インスタンス名の設定確認

           インスタンス名が設定されているかどうか確認する

        Returns:
            named (bool)
        """
        core_model = self.get_core()
        if core_model is not None:
            return core_model.is_named()
        else:
            return self.name is not None
    
    def get_model_name(self):
        """モデル名の取得

           モデル名を取得する。

        Returns:
            model name (str)
        """
        core_model = self.get_core()
        if core_model is not None:
            return core_model.get_model_name()
        else:
            return self.__class__.__name__
    
    def get_info(self, depth :int=0, *, columns: int=70, nest: int=0) -> str:
        """モデル情報取得

           モデルの情報表示用の文字列を取得する
           そのまま表示やログに利用することを想定している

        Returns:
            info (str)
        """
        core_model = self.get_core()
        if core_model is not None:
            return core_model.get_info(depth, columns, nest)

        # セパレータとインデント文字列生成
        indent    = ' ' * (nest*2)
        separetor = '-' * (columns - len(indent))
        name      = self.name
        if name is None:
            name = ''

        # モデルタイトル
        text  = indent + separetor + '\n'
        text += indent + '[' + self.get_model_name() + '] ' + name + '\n'

        # 内容
        text += indent + ' input  shape : ' + str(self.get_input_shape())
        text += indent + ' output shape : ' + str(self.get_output_shape()) + '\n'

        # 最上段なら末尾セパレータ追加
        if nest == 0:
            text +=  indent + separetor + '\n'
        
        return text

    def send_command(self, command, send_to='all'):
        """コマンドの送信

           モデルごとにコマンド文字列で設定を行う
           コマンド文字列の定義はモデルごとに自由である
           Sequentialクラスなどは、保有する下位のモデルに再帰的にコマンドを
           伝搬させるので、複数の層に一括した設定が可能である
           受取先は send_to で送り先はインスタンス名やクラス名で制限でき 'all' を指定すると
           フィルタリングされない

        Args:
            command (str): コマンド文字列
            send_to (str): 送信先
        """
        core_model = self.get_core()
        if core_model is not None:
            self.get_core().send_command(command, send_to)
    
    def set_input_shape(self, input_shape: List[int]):
        """入力シェイプ設定

           BinaryBarainではモデル生成時に入力のシェイプを決定する必要はなく
           ネットワーク構築後に、ネットワークを構成する各モデルの
           set_input_shape を順に呼び出して形状を伝搬させることで
           各モデルの形状の設定を簡易化できる

           set_input_shape が呼ばれるとそれまでの各層で保有する情報は
           保証されない。ネットワーク構築後に一度だけ呼び出すことを想定している
        
        Args:
            input_shape (List[int]): 入力シェイプ

        Returns:
            output_shape (List[int]): 出力シェイプ
        """
#        print('set_input_shape:', self.get_model_name())
#        print('input_shape:', input_shape)

        self.input_shape = input_shape

        core_model = self.get_core()
        if core_model is not None:
            output_shape = self.get_core().set_input_shape(input_shape)
#            print('output_shape:', output_shape)
            return output_shape

        return self.input_shape

    def get_input_shape(self) -> List[int]:
        """入力シェイプ取得

        Returns:
            input_shape (List[int]): 入力シェイプ
        """

        core_model = self.get_core()
        if core_model is not None:
            return core_model.get_input_shape()
        
        return self.input_shape

    def get_output_shape(self) -> List[int]:
        """出力シェイプ取得

        Returns:
            output_shape (List[int]): 出力シェイプ
        """
        core_model = self.get_core()
        if core_model is not None:
            return core_model.get_output_shape()
        
        return self.input_shape

    def get_input_node_size(self) -> int:
        core_model = self.get_core()
        if core_model is not None:
            return core_model.get_input_node_size()
        shape = self.get_input_shape()
        size = 1
        for s in shape:
            size *= s
        return size

    def get_output_node_size(self) -> int:
        core_model = self.get_core()
        if core_model is not None:
            return core_model.get_output_node_size()
        return self.get_input_node_size()

    def get_parameters(self):
        """パラメータ変数取得

           学習対象とするパラメータ群を Variables として返す
           主に最適化(Optimizer)に渡すことを目的としている

           モデル側を自作する際には、このメソッドで戻す変数の中に
           含めなかったパラメータは学習対象から外すことができる為
           パラメータを凍結したい場合などに利用できる

        Returns:
            parameters (Variables): パラメータ変数
        """
        core_model = self.get_core()
        if core_model is not None:
            return bb.Variables.from_core(core_model.get_parameters())
        return bb.Variables()

    def get_gradients(self):
        """勾配変数取得

           get_parameters と対になる勾配変数を Variables として返す
           主に最適化(Optimizer)に渡すことを目的としている

        Returns:
            gradients (Variables): 勾配変数
        """
        core_model = self.get_core()
        if core_model is not None:
            return bb.Variables.from_core(core_model.get_gradients())
        return bb.Variables()
    
    def forward(self, x_buf, train=True):
        """Forward

           モデルは学習および推論の為に forward メソッドを持つ
            train 変数を
        
        Args:
            x_buf (FrameBuffer): 入力データ
            train (bool) : Trueで学習、Falseで推論

        Returns:
            y_buf (FrameBuffer): 出力データ
        """
        core_model = self.get_core()
        if core_model is not None:
            return bb.FrameBuffer.from_core(core_model.forward(x_buf.get_core(), train))
        return x_buf

    def backward(self, dy_buf):
        """Backward

           モデルは学習の為に backword メソッドを持つ
           必ず forwad と対で呼び出す必要があり、直前の forward 結果に対する
           勾配計算を行いながら逆伝搬する

           BinaryBrain は自動微分の機能を備えないので、backword の実装は必須である
        
        Args:
            dy_buf (FrameBuffer): 入力データ

        Returns:
            dx_buf (FrameBuffer): 出力データ
        """
        core_model = self.get_core()
        if core_model is not None:
            return bb.FrameBuffer.from_core(core_model.backward(dy_buf.get_core()))
        return dy_buf

    def dump_bytes(self):
        # バイトデータにシリアライズ(old format)
        core_model = self.get_core()
        if core_model is not None:
            return core_model.dump()
        return b''

    def load_bytes(self, data):
        #バイトデータをロード(old format)
        core_model = self.get_core()
        if core_model is not None:
            size = core_model.load(data)
            return data[size:]
        return data
    
    def __str__(self):
        return self.get_info(1)

    def __repr__(self):
        return self.get_info(1)

    
    
    
        
class Sequential(Model):
    """Sequential class
       複数レイヤーを直列に接続してグルーピングするクラス

       リストの順番で set_input_shape, forward, backward などを実行する
       また send_command の子レイヤーへのブロードキャストや、
       get_parameters, get_gradients の統合を行うことで複数のレイヤーを
       １つのレイヤーとして操作できる

    Args:
        model_list (List[Model]): モデルのリスト
    """
    
    def __init__(self, model_list=None, *, input_shape=None, name=None):
        if model_list is None:
            model_list = []
        self.name = ''
        self.model_list = model_list
        super(Sequential, self).__init__(input_shape=input_shape, name=name)

    def get_core(self):
        # C++のコアの同機能に渡してしまうと Python からの扱いが不便になるので普段はListで管理して必要な時のみ変換する       
        core_model = core.Sequential.create()
        for model in self.model_list:
            core_model.add(model.get_core())
        if self.name is not None:
            core_model.set_name(self.name)            
        return core_model
    
    def set_name(self, name):
        self.name = name

    def get_name(self):
        if self.name is None or len(self.name) == 0:
            return self.get_model_name()
        return self.name

    def is_named(self):
        if self.name is None:
            return False
        if len(self.name) == 0: 
            return False
        return True

    def set_model_list(self, model_list):
        """モデルリストの設定
       
        Args:
            model_list (List[Model]): モデルのリスト
        """
        self.model_list = model_list
    
    def get_model_list(self):
        """モデルリストの取得
       
        Returns:
            model_list (List[Model]): モデルのリスト
        """
        return self.model_list
    
    def __len__(self):
        return len(self.model_list)
    
    def __iter__(self):
        return self.model_list.__iter__()
    
    def __getitem__(self, item):
        return self.model_list[item]
    
    def __setitem__(self, item, model):
        self.model_list[item] = model
    
    def append(self, model):
        """リストへのモデル追加
       
        Args:
            model (Model): リストに追加するモデル
        """
        self.model_list.append(model)
    
    def remove(self, model):
        """リストへからモデル削除
       
        Args:
            model (Model): リストから削除するモデル
        """
        self.model_list.remove(model)


    def get_info(self, depth=0, *, columns=70, nest=0):
        # これ以上ネストしないなら自クラス概要
        if depth > 0 and (nest+1) >= depth:
            return super(Sequential, self).get_info(depth=depth, columns=columns, nest=nest)
        else:
        # セパレータとインデント文字列生成
            indent    = ' ' * (nest*2)
            separetor = '-' * (columns - len(indent))
            name      = self.name
            if name is None:
                name = ''

            # モデルタイトル
            text  = indent + separetor + '\n'
            text += indent + '[' + self.get_model_name() + '] ' + name + '\n'

            # 内容
            text += indent + ' input  shape : ' + str(self.get_input_shape())
            text += indent + ' output shape : ' + str(self.get_output_shape()) + '\n'

            # 子レイヤー
            for model in self.model_list:
                text += model.get_info(depth=depth, columns=columns, nest=nest+1)

            # 最上段なら末尾セパレータ追加
            if nest == 0:
                text +=  indent + separetor + '\n'
            return text

    
    def send_command(self, command, send_to="all"):
        for model in self.model_list:
            model.send_command(command=command, send_to=send_to)
    
    def set_input_shape(self, shape):
        self.input_shape = shape
        for model in self.model_list:
            shape = model.set_input_shape(shape)
        return shape

    def get_parameters(self):
        variables = bb.Variables()
        for model in self.model_list:
            variables.append(model.get_parameters())
        return variables

    def get_gradients(self):
        variables = bb.Variables()
        for model in self.model_list:
            variables.append(model.get_gradients())
        return variables
    
    def forward(self, x_buf, train=True):
        for model in self.model_list:
            x_buf = model.forward(x_buf, train)
        return x_buf

    def backward(self, dy_buf):
        for model in reversed(self.model_list):
            dy_buf = model.backward(dy_buf)
        return dy_buf


    # シリアライズはC++版とフォーマット互換にする
    def dumps(self):
        # ヘッダ
        data = bb.dump_object_header('Sequential')
        
        # バージョン
        ver = 2
        data += bb.int_to_bytes(ver)
        
        # 名前
        name = self.name
        if name is None:
            name = ''
        data += bb.string_to_bytes(name)

        # レイヤ数
        layer_size = len(self.model_list)
        data += bb.int_to_bytes(layer_size)

        # レイヤ本体
        for model in self.model_list:
            data += model.dumps()

        return data

    def loads(self, data):
        # ヘッダ
        data, name = bb.load_object_header(data)
        assert(name == 'Sequential')
        
        # バージョン
        data, ver = bb.int_from_bytes(data)
        assert(ver == 1 or ver == 2)
        
        if ver == 2:
            data, name = bb.string_from_bytes(data)
            if not self.is_named():
                self.name = name

        # レイヤ数
        data, layer_size = bb.int_from_bytes(data)

        # レイヤ本体
        if self.model_list is not None:
            assert(layer_size == len(self.model_list))
            for model in self.model_list:
                data = model.loads(data)
        else:
            self.model_list = []
            for _ in range(layer_size):
                data, model = bb.object_loads(data)
                self.model_list.append(model)
        
        return data

    @classmethod
    def from_bytes(cls, data):
        new_model = cls()
        new_model.model_list=None
        data = new_model.loads(data)
        return data, new_model

model_creator_regist('Sequential', Sequential.from_bytes)



class Switcher(Model):
    """モデル切り替え用基底クラス
    
      主に蒸留や転移学習などでレイヤー差し替えに用いる
      send_command から 'switch_model [name]' でも切り替え可能

    Args:
        model_dict (Dict{str: Model}): 切り替えるモデルの辞書
        init_model_name (str): 初期選択するモデルの名前
    """
    
    def __init__(self, model_dict=None, init_model_name=None, *, input_shape=None, name=None):
        if model_dict is None:
            model_dict = {}
        self.model_dict = model_dict
        self.current_model_name = None
        self.current_model = None
        super(Switcher, self).__init__(input_shape=input_shape, name=name)
        self.switch_model(init_model_name)
    
    def switch_model(self, model_name: str):
        """モデルを切り替える
        
        Args:
            model_dict (Dict{str: Model}): 切り替えるモデルの辞書
        """
        if self.model_dict is not None and model_name is not None:
            if model_name in self.model_dict:
                self.current_model_name = model_name
                self.current_model      = self.model_dict[model_name]
    
    def __getitem__(self, key):
        return self.model_dict[key]

    def __setitem__(self, key, value):
        self.model_dict[key] = value

    def get_model_list(self):
        return self.model_dict.values()
    
    def get_current_model(self):
        return self.current_model

    def get_current_model_name(self):
        return self.current_model_name
    
    def get_core(self):
        current_model = self.get_current_model()
        if current_model is not None:
            return current_model.get_core()
        return None

    def get_info(self, depth :int=0, *, columns: int=70, nest: int=0) -> str:
        current_model = self.get_current_model()
        if current_model is not None:
            return current_model.get_info(depth=depth, columns=columns, nest=nest)
        return ''
    
    def send_command(self, command, send_to='all'):
        # 自分宛なら解釈
        if send_to == 'all' or send_to == self.get_name() or send_to == self.get_model_name():
            args = command.split()
            if len(args) == 2 and args[0] == 'switch_model':
                self.switch_model(args[1])
        
        # 子にも伝搬
        current_model = self.get_current_model()
        if current_model is not None:
            return current_model.send_command(command=command, send_to=send_to)

    def set_input_shape(self, input_shape: List[int]):
        output_shape = None
        for name in self.model_dict:
            shape = self.model_dict[name].set_input_shape(input_shape)
            assert(output_shape is None or shape == output_shape) # すべて同じでなければNG
            output_shape = shape
        return output_shape
    
    def get_input_shape(self) -> List[int]:
        return self.get_current_model().get_input_shape()

    def get_output_shape(self) -> List[int]:
        return self.get_current_model().get_output_shape()

    def get_input_node_size(self) -> int:
        return self.get_current_model().get_input_node_size()

    def get_output_node_size(self) -> int:
        return self.get_current_model().get_output_node_size()

    def get_parameters(self):
        return self.get_current_model().get_parameters()

    def get_gradients(self):
        return self.get_current_model().get_gradients()

    def forward(self, x_buf, train=True):
        return self.get_current_model().forward(x_buf=x_buf, train=train)
    
    def backward(self, dy_buf):
        return self.get_current_model().backward(dy_buf=dy_buf)
    

    # シリアライズは将来C++版フォーマット互換にすることも考慮
    def dumps(self):
        # ヘッダ
        data = b''
        data += core.Object.write_header('Switcher')

        # バージョン
        ver = 1
        data += bb.int_to_bytes(ver)
        
        # 現在のモデル
        data += bb.string_to_bytes(self.current_model_name)

        # レイヤ数
        layer_size = len(self.model_dict)
        data += bb.int_to_bytes(layer_size)

        # レイヤ本体
        for name, model in self.model_dict.items():
            data += bb.string_to_bytes(name)
            data += model.dumps()

        return data

    def loads(self, data):
        # ヘッダ
        load_size, name = core.Object.read_header(data)
        data = data[load_size:]
        assert(name == 'Switcher')
        
        # バージョン
        data, ver = bb.int_from_bytes(data)
        assert(ver == 1)

        # 現在のモデル
        data, model_name = bb.string_from_bytes(data)

        # レイヤ数
        data, layer_size = bb.int_from_bytes(data)

        # レイヤ本体
        if self.model_dict is not None:
            assert(layer_size == len(self.model_dict))
            for _ in range(layer_size):
                data, name = bb.string_from_bytes(data)
                assert(name in self.model_dict)
                data = self.model_dict[name].loads(data)
        else:
            self.model_dict = {}
            for _ in range(layer_size):
                data, name = bb.string_from_bytes(data)
                data, model = bb.object_loads(data)
                self.model_dict[name] = model
        
        # 再構成
        self.switch_model(model_name)

        return data


    # 旧フォーマットは適当に対応
    def dump_bytes(self):
        bytes_dict = {}
        for name in self.model_dict:
            bytes_dict[name] = self.model_dict[name].dump_bytes()
        return pickle.dumps(bytes_dict)
        
    def load_bytes(self, data):
        bytes_dict = pickle.loads(data)
        for name in self.model_dict:
            if name in bytes_dict:
                self.model_dict[name].load_bytes(bytes_dict[name])

    @classmethod
    def from_bytes(cls, data):
        new_model = cls()
        new_model.model_dict=None
        data = new_model.loads(data)
        return data, new_model

model_creator_regist('Switcher', Switcher.from_bytes)



# ------- バイナリ変調 --------

class RealToBinary(Model):
    """RealToBinary class
        実数値をバイナリ値に変換する。
        バイナリ変調機能も有しており、フレーム方向に変調した場合フレーム数(=ミニバッチサイズ)が
        増える。
        またここでビットパッキングが可能であり、32フレームのbitをint32に詰め込みメモリ節約可能である

    Args:
        frame_modulation_size (int): フレーム方向への変調数(フレーム数が増える)
        depth_modulation_size (int): Depth方向への変調数(チャンネル数が増える)
        framewise (bool): Trueで変調閾値をフレーム単位とする(Falseでピクセル単位)
        bin_dtype (DType): 出力の型を bb.DType.FP32 もしくは bb.DType.BIT で指定可能
    """
    
    def __init__(self, *,
                     input_shape=None, frame_modulation_size=1, depth_modulation_size=1, value_generator=None,
                     framewise=False, input_range_lo=0.0, input_range_hi=1.0, name=None,
                     bin_dtype=bb.DType.FP32, real_type=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('RealToBinary', [bin_dtype, real_type]).create
            core_model = core_creator(frame_modulation_size, depth_modulation_size,
                                value_generator, framewise, input_range_lo, input_range_hi)
        super(RealToBinary, self).__init__(core_model=core_model, input_shape=input_shape, name=name)
    

model_creator_regist('RealToBinary', RealToBinary.from_bytes)


class BinaryToReal(Model):
    """BinaryToReal class
        バイナリ値を実数値に戻す。その際にフレーム方向に変調されたデータを積算して
        元に戻すことが可能である

    Args:
        frame_integration_size (int): フレーム方向の積算サイズ数(フレーム変調の統合)
        depth_integration_size (int): チャンネル方向の積算サイズ(0の時はoutput_shape優先)
        output_shape (List[int]): 出力のシェイプ(指定が無ければ入力と同じshape)
        bin_dtype (DType): 入力の型を bb.DType.FP32 もしくは bb.DType.BIT で指定可能
    """
    
    def __init__(self, *, frame_integration_size=1, depth_integration_size=1, output_shape=None, input_shape=None, name=None,
                                                    bin_dtype=bb.DType.FP32, real_type=bb.DType.FP32, core_model=None):
        if output_shape is None:
            output_shape = []
        if core_model is None:
            core_creator = search_core_model('BinaryToReal', [bin_dtype, real_type]).create
            core_model = core_creator(frame_integration_size=frame_integration_size, depth_integration_size=depth_integration_size, output_shape=output_shape)
        
        super(BinaryToReal, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('BinaryToReal', BinaryToReal.from_bytes)



class BitEncode(Model):
    """BitEncode class
        実数値をバイナリ表現としてdepth方向に展開する

    Args:
        bit_size (int): エンコードするbit数
    """

    def __init__(self, bit_size=1, *, output_shape=None, input_shape=None, name=None,
                        bin_dtype=bb.DType.FP32, real_dtype=bb.DType.FP32, core_model=None):
        if output_shape is None:
            output_shape = []
        if core_model is None:
            core_creator = search_core_model('BitEncode', [bin_dtype, real_dtype]).create
            core_model = core_creator(bit_size, output_shape)
        
        super(BitEncode, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('BitEncode', BitEncode.from_bytes)



class Reduce(Model):
    """Reduce class
        多重化されている出力を折り返して積和する

    Args:
        output_shape ([int]): 出力のシェイプ
        integration_size (int): 積算するサイズ
    """

    def __init__(self, output_shape=None, integration_size=0, *, input_shape=None, name=None,
                        bin_dtype=bb.DType.FP32, real_dtype=bb.DType.FP32, core_model=None):
        if output_shape is None:
            output_shape = []
        if core_model is None:
            core_creator = search_core_model('Reduce', [bin_dtype, real_dtype]).create
            core_model = core_creator(output_shape, integration_size)
        
        super(Reduce, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('Reduce', Reduce.from_bytes)



# ------- 演算 --------

class DenseAffine(Model):
    """DenseAffine class
       普通のDenseAffine

    Args:
        output_shape (List[int]): 出力のシェイプ
        initialize_std (float) : 重み初期化乱数の標準偏差
        initializer (str): 変数の初期化アルゴリズム選択。今のところ 'he' のみ
        seed (int): 変数初期値などの乱数シード
    """
    
    def __init__(self, output_shape=None, *, input_shape=None, initialize_std=0.01, initializer='he', seed=1, name=None,
                        dtype=bb.DType.FP32, core_model=None):
        if output_shape is None:
            output_shape = []
        if core_model is None:
            core_creator = search_core_model('DenseAffine', [dtype]).create
            core_model = core_creator(output_shape=output_shape, initialize_std=initialize_std, initializer=initializer, seed=seed)
        
        super(DenseAffine, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('DenseAffine', DenseAffine.from_bytes)


class DepthwiseDenseAffine(Model):
    """DepthwiseDenseAffine class

       Convolution2d と組み合わせて、普通の Depthwise Convolution を作るためのクラス

    Args:
        output_shape (List[int]): 出力のシェイプ
        input_point_size (int): 入力のpoint数
        depth_size (int): depthサイズ
        initialize_std (float) : 重み初期化乱数の標準偏差
        initializer (str): 変数の初期化アルゴリズム選択。今のところ 'he' のみ
        seed (int): 変数初期値などの乱数シード
    """
    
    def __init__(self, output_shape=None, *, input_shape=None, input_point_size=0, depth_size=0,
                            initialize_std=0.01, initializer='he', seed=1, name=None, dtype=bb.DType.FP32, core_model=None):
        if output_shape is None:
            output_shape = []
        if core_model is None:
            core_creator = search_core_model('DepthwiseDenseAffine', [dtype]).create
            core_model = core_creator(output_shape=output_shape, initialize_std=initialize_std, initializer=initializer, seed=seed)
        
        super(DepthwiseDenseAffine, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('DepthwiseDenseAffine', DepthwiseDenseAffine.from_bytes)


class SparseModel(Model):
    """疎結合の基底モデル
    """

    def get_node_connection_size(self, output_node):
        return self.get_core().get_node_connection_size(output_node)

    def set_node_connection_index(self, output_node, connection, input_node):
        self.get_core().set_node_connection_index(output_node, connection, input_node)

    def get_node_connection_index(self, output_node, connection):
        return self.get_core().get_node_connection_index(output_node, connection)

    def get_lut_table_size(self, node):
        return self.get_core().get_lut_table_size(node)

    def get_lut_table(self, node, bitpos):
        return self.get_core().get_lut_table(node, bitpos)


    def get_connection_list(self):
        """接続リスト取得
            
            各出力ノードの入力に対する接続を二次元リストで返す
            入出力のシェイプにかかわらず両者が flatten された状態で処理される

        Returns:
            connection_list (List[List[int]]) : 接続リスト
        """

        connection_list = []
        rows = self.get_output_node_size()
        for i in range(rows):
            node_list = []
            cols = self.get_node_connection_size(i)
            for j in range(cols):
                node_list.append(self.get_node_connection_index(i, j))
            connection_list.append(node_list)
        return connection_list

    def set_connection_list(self, connection_list):
        """接続行列設定
            
            各出力ノードの入力に対する接続を二次元リストで設定する
            入出力のシェイプにかかわらず両者が flatten された状態で2次元リストとする

        Args:
            connection_list (List[List[int]]) : 接続リスト
        """

        input_node_size = self.get_input_node_size()
        rows = self.get_output_node_size()
        assert(len(connection_list) == rows)
        for i in range(rows):
            cols = self.get_node_connection_size(i)
            assert(len(connection_list[i]) == cols)
            for j in range(cols):
                index = int(connection_list[i][j])
                assert(index>= 0 and index < input_node_size)
                self.set_node_connection_index(i, j, index)

    def get_lut_table_list(self):
        """LUTテーブルのリスト取得

        Returns:
            lut_list (List[List[int]]) : 接続行列
        """

        lut_list = []
        rows = self.get_output_node_size()
        for i in range(rows):
            bit_list = []
            cols = self.get_lut_table_size(i)
            for j in range(cols):
                bit_list.append(self.get_lut_table(i, j))
            lut_list.append(bit_list)
        return lut_list


class DifferentiableLut(SparseModel):
    """DifferentiableLut class
        微分可能LUTモデル

        内部計算的には StocasticLUT + BatchNormalization + Binarize(HardTanh) で構成される

        FPGA合成するためのルックアップテーブル型のモデルを学習することができる
        純粋な Stochastic 演算のみを行いたい場合は binarize と batch_norm の両方を False にすればよい。
        
    Args:
        output_shape (List[int]): 出力のシェイプ
        connection(str): 結線ルールを 'random', 'serial', 'depthwise' から指定可能
        batch_norm (bool): BatchNormalization を有効にするか
        momentum (float): BatchNormalization を有効にするか
        gamma (float): BatchNormalization を有効にするか
        beta (float): BatchNormalization を有効にするか
        N (int): LUTの入力数
        seed (int): 変数初期値などの乱数シード
        bin_dtype (DType)): バイナリ出力の型を bb.DType.FP32 と bb.DType.BIT から指定(bb.DType.BIT は binarize=True 時のみ)
    """
    
    def __init__(self, output_shape=None, *, input_shape=None,
                    connection='random', binarize=True, batch_norm=True, momentum=0.0, gamma= 0.3, beta=0.5, seed=1,
                    name=None, N=6, bin_dtype=bb.DType.FP32, real_dtype=bb.DType.FP32, core_model=None):
        if output_shape is None:
            output_shape = []
        if core_model is None:
            core_creator = search_core_model('DifferentiableLut' + str(N), [bin_dtype, real_dtype]).create
            core_model = core_creator(output_shape, batch_norm, binarize, connection, momentum, gamma, beta, seed)

        super(DifferentiableLut, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

    def W(self):
        """重み行列取得
        
            コピーではなくスライス参照を得ており、本体の値を書き換え可能

        Returns:
            W (Tensor): 重み行列を指すTensor
        """
        return bb.Tensor.from_core(self.get_core().W())
    
    def dW(self):
        """重みの勾配行列取得

            コピーではなくスライス参照を得ており、本体の値を書き換え可能

        Returns:
            W (Tensor): 重みの勾配を指すTensor
        """
        return bb.Tensor.from_core(self.get_core().dW())

    def get_mean(self):
        return bb.Tensor.from_core(self.get_core().get_mean())
    
    def get_var(self):
        return bb.Tensor.from_core(self.get_core().get_var())
    
    def get_gamma(self):
        return self.get_core().get_gamma()
    
    def get_beta(self):
        return self.get_core().get_beta()

model_creator_regist('DifferentiableLut6', DifferentiableLut.from_bytes)
model_creator_regist('DifferentiableLut5', DifferentiableLut.from_bytes)
model_creator_regist('DifferentiableLut4', DifferentiableLut.from_bytes)
model_creator_regist('DifferentiableLut3', DifferentiableLut.from_bytes)
model_creator_regist('DifferentiableLut2', DifferentiableLut.from_bytes)



class BinaryLut(SparseModel):
    """バイナリLUTモデル

       一般的なFPGAのLUTと同等の機能をするモデル。
       学習能力はなく、他のモデルで学習した結果をインポートしてモデル評価を行うためのもの
        
    Args:
        output_shape (List[int]): 出力のシェイプ
        connection(str): 結線ルールを 'random', 'serial', 'depthwise' から指定可能(未実装)
        N (int): LUTの入力数
        seed (int): 変数初期値などの乱数シード
        fw_dtype (DType)): forwardの型を bb.DType.FP32 と bb.DType.BIT から指定
    """
    
    def __init__(self, output_shape=None, *, input_shape=None,
                    connection='random', seed=1, name=None, N=6, fw_dtype=bb.DType.FP32, bw_dtype=bb.DType.FP32, core_model=None):
        if output_shape is None:
            output_shape = []
        if core_model is None:
            core_creator = search_core_model('BinaryLut' + str(N), [fw_dtype, bw_dtype]).create
            core_model  = core_creator(output_shape, connection, seed)
        
        super(BinaryLut, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

    def import_layer(self, leyaer):
        """他のモデルからからインポート

            インポート元は入出力が同じ形状をしている必要があり、デジタル値の入力に対して
            出力を 0.5 を閾値として、バイナリテーブルを構成して取り込む

        Args:
            leyaer (Model): インポート元のモデル
        """
        self.get_core().import_layer(leyaer.get_core())

    @staticmethod
    def from_sparse_model(layer, *, fw_dtype=bb.DType.FP32):
        """他のモデルを元に生成

            インポート元は入出力が同じ形状をしている必要があり、デジタル値の入力に対して
            出力を 0.5 を閾値として、バイナリテーブルを構成して取り込む

        Args:
            leyaer (Model): インポート元のモデル
        """
        N = layer.get_core().get_node_connection_size(0)
        new_model = BinaryLut(output_shape=layer.get_output_shape(), input_shape=layer.get_input_shape(),
                                 N=N, fw_dtype=fw_dtype)
        new_model.import_layer(layer)
        return new_model


model_creator_regist('BinaryLut6', BinaryLut.from_bytes)
model_creator_regist('BinaryLut5', BinaryLut.from_bytes)
model_creator_regist('BinaryLut4', BinaryLut.from_bytes)
model_creator_regist('BinaryLut3', BinaryLut.from_bytes)
model_creator_regist('BinaryLut2', BinaryLut.from_bytes)
model_creator_regist('BinaryLut1', BinaryLut.from_bytes)



# ------- フィルタ --------


class ConvolutionIm2Col(Model):
    """ConvolutionIm2Col class
       畳み込みの lowering における im2col 層
    """

    def __init__(self, filter_size=(1, 1), stride=(1, 1), *,
                        padding='valid', border_mode='reflect_101', border_value=0.0,
                        input_shape=None, name=None, fw_dtype=bb.DType.FP32, bw_dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('ConvolutionIm2Col', [fw_dtype, bw_dtype]).create
            core_model = core_creator(filter_size[0], filter_size[1], stride[0], stride[1], padding, border_mode)
            
        super(ConvolutionIm2Col, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

    def get_filter_size(self):
        core_model = self.get_core()
        return [core_model.get_filter_size_h(), core_model.get_filter_size_w()]

    def get_stride(self):
        core_model = self.get_core()
        return [core_model.get_stride_y(), core_model.get_stride_x()]

    def get_padding(self):
        return self.get_core().get_padding()

    def get_border_mode(self):
        return self.get_core().get_border_mode()

    def get_border_value(self):
        return self.get_core().get_border_value()

model_creator_regist('ConvolutionIm2Col', ConvolutionIm2Col.from_bytes)



class ConvolutionCol2Im(Model):
    """ConvolutionCol2Im class
       畳み込みの lowering における col2im 層
    """

    def __init__(self, output_size=(1, 1), *, input_shape=None, name=None,
                    fw_dtype=bb.DType.FP32, bw_dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('ConvolutionCol2Im', [fw_dtype, bw_dtype]).create
            core_model = core_creator(output_size[0], output_size[1])
        
        super(ConvolutionCol2Im, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

    def set_output_size(self, output_size):
        self.get_core().set_output_size(output_size[0], output_size[1])

model_creator_regist('ConvolutionCol2Im', ConvolutionCol2Im.from_bytes)



class Convolution2d(Sequential):
    """Convolution class
        Lowering による畳み込み演算をパッキングするクラス

        sub_layer で指定した演算レイヤーを畳み込み計算するためのモデル
        例えば sub_layer に DenseAffineレイヤーを指定すると一般的なCNN用の畳み込み層となるが、
        BinaryBrain ではここに DifferentiableLut モデルを組み合わせて作った複合レイヤーを
        指定することでFPGA化に適した畳み込み層を学習させることができる。

        sub_layer で指定したサブレイヤーが im2col と col2im で挟み込まれ、一般に
        Lowering と呼ばれる方法で畳み込み演算が実行される

    Args:
        sub_layer (Model): 畳み込みを行うサブレイヤー(このレイヤが im2col と col2im で挟み込まれる)
        filter_size ((int, int)): 2次元のタプルでフィルタサイズを指定する
        stride ((int, int)): 2次元のタプルでストライドサイズを指定する
        batch_norm (bool): BatchNormalization を有効にするか
        padding (str)): パディングの方法を 'valid' と 'same' で指定する
        border_mode (Border)): 'same' 時のボーダー処理を指定する
        border_value (float): 'same' 時のボーダー処理が CONSTANT の場合にパディング値を指定する
        fw_dtype (DType)): forwarする型を bb.DType.FP32 と bb.DType.BIT から指定
    """

    def __init__(self, sub_layer, filter_size=(1, 1), stride=(1, 1), *, input_shape=None,
                        padding='valid', border_mode='reflect_101', border_value=0.0,
                        name=None, fw_dtype=bb.DType.FP32, bw_dtype=bb.DType.FP32):
        
#       self.name         = name
#       self.input_shape  = input_shape
#       self.filter_size  = filter_size
#       self.stride       = stride
#       self.padding      = padding
#       self.border_mode  = border_mode
#       self.border_value = border_value
        self.fw_dtype     = fw_dtype
        self.bw_dtype     = bw_dtype
        
        self.im2col       = ConvolutionIm2Col(filter_size=filter_size, stride=stride,
                                padding=padding, border_mode=border_mode, border_value=border_value,
                                fw_dtype=fw_dtype, bw_dtype=bw_dtype)
        self.sub_layer    = sub_layer
        self.col2im       = ConvolutionCol2Im(fw_dtype=fw_dtype, bw_dtype=bw_dtype)
        model_list = [self.im2col, self.sub_layer, self.col2im]

        super(Convolution2d, self).__init__(model_list=model_list, name=name, input_shape=input_shape)

    def send_command(self, command, send_to="all"):
        self.im2col.send_command(command=command, send_to=send_to)
        self.sub_layer.send_command(command=command, send_to=send_to)
        self.col2im.send_command(command=command, send_to=send_to)
    
    def get_core(self):
        core_creator = search_core_model('Convolution2d', [self.fw_dtype, self.bw_dtype]).create
        core_model = core_creator(self.im2col.get_core(), self.sub_layer.get_core(), self.col2im.get_core())
        if self.name is not None:
            core_model.set_name(self.name)
        return core_model

    def get_sub_layer(self):
        return self.sub_layer
    
    def set_input_shape(self, shape):
        self.input_shape = shape
        
        # 出力サイズ計算
        # input_c_size = shape[0]
        input_h_size = shape[1]
        input_w_size = shape[2]
        padding     = self.im2col.get_padding()
        filter_size = self.im2col.get_filter_size()
        stride      = self.im2col.get_stride()
        if padding == "valid":
            output_h_size = ((input_h_size - filter_size[0] + 1) + (stride[0] - 1)) // stride[0]
            output_w_size = ((input_w_size - filter_size[1] + 1) + (stride[1] - 1)) // stride[1]
        elif padding == "same":
            output_h_size = (input_h_size + (stride[0] - 1)) // stride[0]
            output_w_size = (input_w_size + (stride[1] - 1)) // stride[0]
        else:
            raise ValueError("illegal padding value")
        
        self.col2im.set_output_size(output_size=[output_h_size, output_w_size])
        
        super(Convolution2d, self).set_model_list([self.im2col, self.sub_layer, self.col2im])
        
        return super(Convolution2d, self).set_input_shape(shape)
    
    def get_object_name(self):
        return 'Convolution2d_' + bb.dtype_to_name(self.fw_dtype) + '_' + bb.dtype_to_name(self.bw_dtype)

    # シリアライズはC++版とフォーマット互換にする
    def dumps(self):
        # ヘッダ
        data = b''
        data += core.Object.write_header(self.get_object_name())

        # バージョン
        ver = 1
        data += bb.int_to_bytes(ver)
        
        # メンバ
        data += self.im2col.dumps()
        data += self.col2im.dumps()
        
        # 子レイヤー
        if self.sub_layer:
            data += bb.bool_to_bytes(True)
            data += self.sub_layer.dumps()
        else:
            data += bb.bool_to_bytes(False)

        return data
    
    def loads(self, data):
        # ヘッダ
        data, name = bb.load_object_header(data)
        type_names = re.match('Convolution2d_(.+)_(.+)', name)
        assert(type_names)
        self.fw_dtype = bb.dtype_from_name(type_names[1])
        self.bw_dtype = bb.dtype_from_name(type_names[2])
        
        # バージョン
        data, ver = bb.int_from_bytes(data)
        assert(ver == 1)

        # メンバ
        data = self.im2col.loads(data)
        data = self.col2im.loads(data)
        
        # 子レイヤー
        data, has_layer = bb.bool_from_bytes(data)

        # レイヤ本体
        if has_layer:
            if self.sub_layer:
                data = self.sub_layer.loads(data)
            else:
                data, model = bb.object_loads(data)
                self.sub_layer = model
        
        # 再構成
        self.set_model_list([self.im2col, self.sub_layer, self.col2im])

        return data

    @classmethod
    def from_bytes(cls, data):
        _, object_name = bb.load_object_header(data)
        dtypes = object_name.split('_')
        new_model = cls(sub_layer=None, fw_dtype=bb.dtype_from_name(dtypes[1]), bw_dtype=bb.dtype_from_name(dtypes[2]))
        data = new_model.loads(data)
        return data, new_model

model_creator_regist('Convolution2d', Convolution2d.from_bytes)


class MaxPooling(Model):
    """MaxPooling class

    Args:
        filter_size ((int, int)): 2次元のタプルでフィルタサイズを指定する
        fw_dtype (DType)): forwarする型を bb.DType.FP32 と bb.DType.BIT から指定
    """

    def __init__(self, filter_size=(2, 2), *, input_shape=None, name=None,
                    fw_dtype=bb.DType.FP32, bw_dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('MaxPooling', [fw_dtype, bw_dtype]).create
            core_model = core_creator(filter_size[0], filter_size[1])
        
        super(MaxPooling, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('MaxPooling', MaxPooling.from_bytes)



class StochasticMaxPooling(Model):
    """StochasticMaxPooling class

        Stochastic 演算として OR 演算で Pooling 層を構成するモデル

    Args:
        filter_size ((int, int)): 2次元のタプルでフィルタサイズを指定する(現在2x2のみ)
        fw_dtype (DType)): forwarする型を bb.DType.FP32 と bb.DType.BIT から指定
    """

    def __init__(self, filter_size=(2, 2), *, input_shape=None, name=None,
                    fw_dtype=bb.DType.FP32, bw_dtype=bb.DType.FP32, core_model=None):
        assert(len(filter_size)==2)

        if core_model is None:
            if  filter_size[0]==2 and filter_size[1]==2:
                core_creator = search_core_model('StochasticMaxPooling2x2', [fw_dtype, bw_dtype]).create
                core_model = core_creator()
            else:
                core_creator = search_core_model('StochasticMaxPooling', [fw_dtype, bw_dtype]).create
                core_model = core_creator(filter_size[0], filter_size[1])

        super(StochasticMaxPooling, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('StochasticMaxPooling2x2', StochasticMaxPooling.from_bytes)
model_creator_regist('StochasticMaxPooling', StochasticMaxPooling.from_bytes)


class UpSampling(Model):
    """UpSampling class

        畳み込みの逆方向にアップサンプリングを行うモデル

    Args:
        filter_size ((int, int)): 2次元のタプルでフィルタサイズを指定する(現在2x2のみ)
        fw_dtype (DType)): forwarする型を bb.DType.FP32 と bb.DType.BIT から指定
    """

    def __init__(self, filter_size=(2, 2), *, fill=True, input_shape=None, name=None,
                        fw_dtype=bb.DType.FP32, bw_dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('UpSampling', [fw_dtype, bw_dtype]).create
            core_model = core_creator(filter_size[0], filter_size[1], fill)

        super(UpSampling, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('UpSampling', UpSampling.from_bytes)


# ------- 活性化 --------

class Binarize(Model):
    """Binarize class

    2値化(活性化層)
    backward は hard-tanh となる

    Args:
        bin_dtype (DType)): バイナリ型を bb.DType.FP32 と bb.DType.BIT から指定
    """

    def __init__(self, *, input_shape=None, name=None, bin_dtype=bb.DType.FP32, real_dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('Binarize', [bin_dtype, real_dtype]).create
            core_model = core_creator()

        super(Binarize, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('Binarize', Binarize.from_bytes)


class Sigmoid(Model):
    """Sigmoid class

       Sigmoid 活性化層
       send_command で "binary true" とすることで、Binarize に切り替わる
       多値で学習を進めて、途中から Binarize に切り替える実験などが可能である
    """
    def __init__(self, *, input_shape=None, name=None, bin_dtype=bb.DType.FP32, real_dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('Sigmoid', [bin_dtype, real_dtype]).create
            core_model = core_creator()

        super(Sigmoid, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('Sigmoid', Sigmoid.from_bytes)


class ReLU(Model):
    """ReLU class

       ReLU 活性化層
       send_command で "binary true" とすることで、Binarize に切り替わる
       多値で学習を進めて、途中から Binarize に切り替える実験などが可能である
    """
    def __init__(self, *, input_shape=None, name=None, bin_dtype=bb.DType.FP32, real_dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('ReLU', [bin_dtype, real_dtype]).create
            core_model = core_creator()

        super(ReLU, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('ReLU', ReLU.from_bytes)


class HardTanh(Model):
    """HardTanh class

       HardTanh 活性化層
       send_command で "binary true" とすることで、Binarize に切り替わる
       多値で学習を進めて、途中から Binarize に切り替える実験などが可能である
    """
    def __init__(self, *, input_shape=None, name=None, bin_dtype=bb.DType.FP32, real_dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('HardTanh', [bin_dtype, real_dtype]).create
            core_model = core_creator()

        super(HardTanh, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('HardTanh', HardTanh.from_bytes)



class Softmax(Model):
    """Softmax class

       Softmax 活性化層
    """
    def __init__(self, *, input_shape=None, name=None, dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('Softmax', [dtype]).create
            core_model = core_creator()

        super(Softmax, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('Softmax', Softmax.from_bytes)


# ------- 補助モデル --------

class BatchNormalization(Model):
    """BatchNormalization class

    Args:
        momentum (float): 学習モーメント
        gamma (float): gamma 初期値
        beta (float): beta 初期値
        fix_gamma (bool): gamma を固定する(学習させない)
        fix_beta (bool): beta を固定する(学習させない)
        bin_dtype (DType)): バイナリ型を bb.DType.FP32 と bb.DType.BIT から指定
    """

    def __init__(self, *, input_shape=None, momentum=0.9, gamma=1.0, beta=0.0,
                        fix_gamma=False, fix_beta=False, name=None, dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('BatchNormalization', [dtype]).create
            core_model = core_creator(momentum, gamma, beta, fix_gamma, fix_beta)

        super(BatchNormalization, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('BatchNormalization', BatchNormalization.from_bytes)


class Dropout(Model):
    """Dropout class

    Args:
        rate (float): Drop率
        seed (int): 乱数シード
        fw_dtype (DType): forwardの型を bb.DType.FP32 と bb.DType.BIT から指定
    """

    def __init__(self, *, rate=0.5, input_shape=None, seed=1, name=None,
                    fw_dtype=bb.DType.FP32, bw_dtype=bb.DType.FP32, core_model=None):
        if core_model is None:
            core_creator = search_core_model('Dropout', [fw_dtype, bw_dtype]).create
            core_model = core_creator(rate, seed)

        super(Dropout, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('Dropout', Dropout.from_bytes)


class Shuffle(Model):
    """Shuffle class

        所謂 ShuffleNet のようなシャッフルを行うモデル
        入力ノードが shuffle_unit 個のグループに分割されるようにシャッフルする


    Args:
        shuffle_unit (int): シャッフルする単位
    """

    def __init__(self, shuffle_unit, *, output_shape=None, input_shape=None, name=None, core_model=None):
        if output_shape is None:
            output_shape = []
        if core_model is None:
            core_creator = search_core_model('Shuffle', []).create
            core_model = core_creator(shuffle_unit, output_shape)

        super(Shuffle, self).__init__(core_model=core_model, input_shape=input_shape, name=name)

model_creator_regist('Shuffle', Shuffle.from_bytes)



# ------- その他 --------


def get_model_list(net, flatten:bool =False):
    ''' Get model list from networks
        ネットから構成するモデルのリストを取り出す
    
        Args:
            net     (Model): 検索するパス
            flatten (bool): 階層をフラットにするかどうか
        Returns:
            list of models
    '''
    
    if  type(net) is not list:
        net = [net]
    
    if not flatten:
        return net
    
    def flatten_list(in_list, out_list):
        for model in in_list:
            if hasattr(model, 'get_model_list'):
                flatten_list(model.get_model_list(), out_list)
            else:
                out_list.append(model)
    
    out_list = []
    flatten_list(net, out_list)
    
    return out_list


def get_model_list_for_rtl(net):
    # RTL化用リストを得る
    
    if  type(net) is not list:
        net = [net]
    
    def flatten_list(in_list, out_list):
        for model in in_list:
            if issubclass(type(model), Switcher):
                model = model.get_current_model()
            if type(model) != Convolution2d and hasattr(model, 'get_model_list'):
                flatten_list(model.get_model_list(), out_list)
            else:
                out_list.append(model)
    
    out_list = []
    flatten_list(net, out_list)
    
    return out_list



# -*- coding: utf-8 -*-

import binarybrain      as bb
import binarybrain.core as core
import numpy as np
from typing import List



class Optimizer(bb.Object):
    """Optimizer の基本クラス
    """
    
    def __init__(self, core_optimizer=None):
        super(Optimizer, self).__init__(core_object=core_optimizer)
    
    def set_variables(self, params, grads):
        """変数設定

        Args:
            params (Variables): 学習対象のパラメータ変数
            grads (Variables): paramsに対応する勾配変数
        """

        self.get_core().set_variables(params.get_core(), grads.get_core())
    
    def update(self):
        """パラメータ更新

            set_variablesで設定された勾配変数に基づいた学習をset_variablesで
            設定されたパラメータ変数に適用する
        """

        return self.get_core().update()


class OptimizerSgd(Optimizer):
    """SGD 最適化クラス

    Args:
        learning_rate (float): 学習率
    """
    def __init__(self, learning_rate=0.01, dtype=bb.DType.FP32):
        core_optimizer = bb.search_core_object('OptimizerSgd', [dtype]).create()
        super(OptimizerSgd, self).__init__(core_optimizer=core_optimizer)


class OptimizerAdaGrad(Optimizer):
    """AdaGrad 最適化クラス

    Args:
        learning_rate (float): 学習率
    """

    def __init__(self, learning_rate=0.01, dtype=bb.DType.FP32):
        core_optimizer = bb.search_core_object('OptimizerAdaGrad', [dtype]).create()
        super(OptimizerAdaGrad, self).__init__(core_optimizer=core_optimizer)


class OptimizerAdam(Optimizer):
    """Adam 最適化クラス

    Args:
        learning_rate (float): 学習率
        beta1 (float): beta1
        beta2 (float): beta2
    """

    def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999, dtype=bb.DType.FP32):
        core_optimizer = bb.search_core_object('OptimizerAdam', [dtype]).create()
        super(OptimizerAdam, self).__init__(core_optimizer=core_optimizer)





# -*- coding: utf-8 -*-

import binarybrain      as bb
import binarybrain.core as core


def get_version_string():
    """バージョン文字列取得
    
    Returns:
        version (str) : バージョン文字列
    """
    return core.get_version_string()


def omp_set_num_threads(threads: int):
    """omp_set_num_threads
    
       omp_set_num_threadsを呼び出す
       バックグランドで学習する場合など、Host側のCPUをすべて使うと
       逆に性能が落ちる場合や、運用上不便なケースなどで個数制限できる

    Args:
        threads (int) : OpenMPでのスレッド数
    """
    core.omp_set_num_threads(threads)


def is_device_available():
    """デバイス(GPU)が有効かの確認

    Returns:
        device_available (bool) : デバイス(GPU)が利用可能なら True を返す
    """
    return core.Manager.is_device_available()


def set_host_only(host_only: bool):
    """ホスト(CPU)のみの指定

        True を設定するとデバイス(GPU)を未使用としてホスト(CPU)のみを利用

    Args:
        host_only (bool) : ホストのみの場合 True を指定
    """
    core.Manager.set_host_only(host_only)

def get_device_count():
    """利用可能なデバイス(GPU)の個数を確認

    Returns:
        device_count (int) : 利用可能なデバイス(GPU)の個数を返す
    """
    return core.get_device_count()

def set_device(device_id):
    """利用するデバイス(GPU)を切り替え

    Args:
        device_id (int) : 利用するデバイス番号を指定
    """
    core.set_device(device_id)

def get_device_properties_string(device_id):
    """現在のデバイス(GPU)の情報を入れた文字列を取得

    Args:
        device_id (int) : 情報を取得するデバイス番号を指定

    Returns:
        device_properties_string (str) : 現在のデバイス(GPU)の情報を入れた文字列を返す
    """
    return core.get_device_properties_string(device_id)


def get_device_properties(device_id):
    return core.get_device_properties(device_id)

def get_device_name(device_id):
    return core.get_device_name(device_id)


def get_device_allocated_memory_size():
    return core.get_device_allocated_memory_size()

def garbage_collect_device_memory():
    return core.garbage_collect_device_memory()



# -*- coding: utf-8 -*-

import os
import datetime
import glob
import re
import shutil
import pickle

import binarybrain as bb


def get_date_string():
    # データ保存パス用の日付文字列を生成
    return datetime.datetime.now().strftime('%Y%m%d_%H%M%S')

def is_date_string(text: str):
    ''' Check if the string is a date
        データ保存パス用の日付文字列かどうか判定
    
        Args:
            text (str): 判定する文字列
        
        Returns:
            Boolean.
    '''
    
    if re.fullmatch('[12][0-9]{3}[01][0-9][0-3][0-9]_[0-2][0-9][0-5][0-9][0-5][0-9]', text):
        return True
    return False
    
def get_latest_path(path: str) -> str:
    ''' Get latest data path
        最新のデータ保存パスを取得
    
        Args:
            path (str): 検索するパス
        
        Returns:
            見つかったパス. 見つからなければ None
    '''
    
    if not os.path.exists(path):
        return None
    
    files = os.listdir(path)
    dirs  = [f for f in files if os.path.isdir(os.path.join(path, f))]
    
    targets = []
    for d in dirs:
        if is_date_string(d):
            targets.append(d)
    
    if not targets:
        return None
    
    targets.sort(reverse=True)
    return os.path.join(path, targets[0])


def remove_old(path: str, keeps: int=-1):
    ''' Get latest data path
        最新のデータ保存パスを取得
    
        Args:
            path (str): 検索するパス
            keeps (int): 削除せずに残す数
    '''
    
    if keeps < 0:
        return
    
    files = os.listdir(path)
    dirs  = [f for f in files if os.path.isdir(os.path.join(path, f))]
    
    targets = []
    for d in dirs:
        if is_date_string(d):
            no_delete_file = os.path.join(path, d, '__no_delete__')
            if not os.path.exists(no_delete_file):
                targets.append(d)
    
    targets.sort(reverse=True)
    del targets[:keeps]
    
    for t in targets:
        shutil.rmtree(os.path.join(path, t))


def _save_net_file(path: str, name: str, net, file_format=None):
    if file_format == 'bin' or file_format=='all':
        # 旧バージョン
        net_file_name = os.path.join(path, name + '.bin')
        with open(net_file_name, 'wb') as f:
            f.write(net.dump_bytes())
    
    elif file_format == 'pickle' or file_format=='all':
        # pickle
        net_file_name = os.path.join(path, name + '.pickle')
        with open(net_file_name, 'wb') as f:
            f.write(pickle.dumps(net))
    else:
        # デフォルトフォーマット
        net_file_name = os.path.join(path, name + '.bb_net')
        with open(net_file_name, 'wb') as f:
            f.write(net.dumps())

def _load_net_file(path: str, name: str, net, file_format=None) -> bool:
    # デフォルトフォーマット
    if file_format is None or file_format == 'bb_net':
        net_file_name = os.path.join(path, name + '.bb_net')
        if os.path.exists(net_file_name):
            with open(net_file_name, 'rb') as f:
                net.loads(f.read())
            return True
    
    # 無ければ旧フォーマットを探してみる
    if file_format is None or file_format == 'bin':
        net_file_name = os.path.join(path, name + '.bin')
        if os.path.exists(net_file_name):
            with open(net_file_name, 'rb') as f:
                net.load_bytes(f.read())
            return True

    # pickle 
    if file_format is None or file_format == 'pickle':
        net_file_name = os.path.join(path, name + '.pickle')
        if os.path.exists(net_file_name):
            with open(net_file_name, 'rb') as f:
                tmp_net = pickle.loads(f.read())
            # pickle はインスタンスが作り直されてしまうのでコピー
            net.loads(tmp_net.dumps())
            return True

    return False


def save_models(path: str, net, *, write_layers=True, file_format=None):
    ''' save networks
        ネットを構成するモデルの保存
        
        Args:
            path (str):  保存するパス
            net (Model): 保存するネット
            write_layers (bool) : レイヤー別にも出力するかどうか
    '''
    
    # make dir
    os.makedirs(path, exist_ok=True)
    
    # save
    net_name = net.get_name()
    _save_net_file(path, net_name, net, file_format=file_format)

    # save flatten models
    if write_layers:
        models = bb.get_model_list(net, flatten=True)
        fname_list = []  # 命名重複回避用
        for i, model in enumerate(models):
            name = model.get_name()
            if model.is_named():
                if name in fname_list:
                    print('[warrning] duplicate model name : %s' % (name))
                    fname = '%04d_%s' % (i, name)
                else:
                    fname = '%s' % (name)
            else:
                fname = '%04d_%s' % (i, name)
            fname_list.append(fname)
            
            _save_net_file(path, fname, model, file_format=file_format)
            


def load_models(path: str, net, *, read_layers: bool=False, file_format=None):
    ''' load networks
        ネットを構成するモデルの保存
        
        Args:
            path (str):  読み出すパス
            net (Model): 読み込むネット
            read_layers (bool) : レイヤー別に読み込むか
    '''

    # load
    if not read_layers:
        net_name = net.get_name()
        res = _load_net_file(path, net_name, net, file_format=file_format)
        if not res:
            print('file not found : %s'%os.path.join(path, net_name))
        return res
    
    # load models
    models    = bb.get_model_list(net, flatten=True)
    fname_list = []
    for i, model in enumerate(models):
        name = model.get_name()
        if model.is_named():
            if name in fname_list:
                print('[warrning] duplicate model name : %s' % (name))
                fname = '%04d_%s' % (i, name)
            else:
                fname = '%s' % (name)
        else:
            fname = '%04d_%s' % (i, name)
        fname_list.append(fname)
        
        res = _load_net_file(path, fname, model, file_format=file_format)
        if not res:
            print('file not found : %s' % fname)
    
    return True


def save_networks(path: str, net, name=None, *, backups: int=10, write_layers: bool=False, file_format=None):
    ''' save networks
        ネットを構成するモデルの保存
        
        指定したパスの下にさらに日付でディレクトリを作成して保存
        古いものから削除する機能あり
        名前を指定すれば日付ではなく名前で記録可能
        
        Args:
            path (str) : 保存するパス
            net (Model) : 保存するネット
            name (str)  : 保存名(指定しなければ日時で保存)
            backups (int) : 残しておく古いデータ数(-1ですべて残す)
            write_layers(bool) : レイヤー別に出力
        
        Returns:
            name (str) : 保存名を返す
    '''
    
    # make dir
    os.makedirs(path, exist_ok=True)
    
    # save with date
    if name is None:
        name = get_date_string()
    data_path = os.path.join(path, name)
    
    save_models(data_path, net, write_layers=write_layers, file_format=file_format)
    
    if backups >= 0:
        remove_old(path, keeps=backups)
    
    return name

def load_networks(path: str, net, name=None, *, read_layers: bool=False, file_format=None):
    ''' load network
        ネットを構成するモデルの読み込み
        
        最新のデータを探して読み込み
        名前を指定した場合はそれを読み込み
        
        Args:
            path (str) : 読み込むパス
            net (Model) : 読み込むネット
            file_format (str) : 読み込む形式(Noneがデフォルト)
    '''
    
    if name is None:
        data_path = get_latest_path(path)
    else:
        data_path = os.path.join(path, name)
    
    if data_path is None:
        print('not loaded : file not found')
        return False
    
    ret = load_models(data_path, net, read_layers=read_layers, file_format=None)
#   print('load : %s' % data_path)

    return ret


def get_load_networks_path(path: str, net, name=None):
    '''ネットワークをロードするパスを取得'''
    if name is None:
        return get_latest_path(path)
    else:
        return os.path.join(path, name)



# -*- coding: utf-8 -*-

import pickle
import numpy as np
import inspect
from typing import List

import binarybrain      as bb
import binarybrain.core as core


# ---- シリアライズ時のデータフォーマット定義 ----

def int_to_bytes(value: int):
    return value.to_bytes(8, 'little')

def int_from_bytes(data: bytes) -> (bytes, int):
    value = int.from_bytes(data[0:8], 'little')
    return data[8:], value

def bool_to_bytes(value: bool):
    return value.to_bytes(1, 'little')

def bool_from_bytes(data: bytes) -> (bytes, bool):
    value = bool.from_bytes(data[0:1], 'little')
    return data[1:], value


def string_to_bytes(value: str):
    value = value.encode(encoding='utf-8')
    data = int_to_bytes(len(value))
    data += value
    return data

def string_from_bytes(data: bytes) -> (bytes, str):
    data, str_len = int_from_bytes(data)
    value = data[0:str_len].decode(encoding='utf-8')
    return data[str_len:], value


def dump_object_header(object_name):
    return core.Object.write_header(object_name)

def load_object_header(data):
    load_size, object_name = core.Object.read_header(data)
    return data[load_size:], object_name



# ---- core のクラス一覧取得管理 ----

_core_class_list = inspect.getmembers(bb.core, inspect.isclass)
_core_class_dict = {k: v for (k, v) in _core_class_list}

def get_core_class_list():
    return _core_class_list

def get_core_class_dict():
    return _core_class_dict

def search_core_class(class_name):
    return _core_class_dict[class_name]


def get_core_subclass_list(superclass):
    class_list = []
    for c in get_core_class_list():
        if issubclass(c[1], superclass):
            class_list.append(c)
    return class_list

def get_core_subclass_dict(superclass):
    return {k: v for (k, v) in get_core_subclass_list(superclass)}


# ---- core.Object ----

def make_core_object_name(object_name, dtypes):
    # core側のオブジェクト名を生成 (オジェクト名＋テンプレート型)
    for dtype in dtypes:
        object_name = object_name + '_' + bb.dtype_to_name(dtype)
    return object_name

def split_core_object_name(core_object_name):
    # core側のオブジェクト名を分解
    args = core_object_name.split('_')
    dtypes = []
    for dtype_name in args[1:]:
        dtypes.append(bb.dtype_from_name(dtype_name))
    return args[0], dtypes


# core の Object class を管理
_core_object_list = get_core_subclass_list(core.Object)
_core_object_dict = get_core_subclass_dict(core.Object)

def get_core_object_list():
    return _core_object_list

def get_core_object_dict():
    return _core_object_dict

def search_core_object(object_name, dtypes):
    core_object_name = make_core_object_name(object_name, dtypes)
    return _core_object_dict[core_object_name]



# ---- core.Object 再生成 ----

def core_object_loads(data):
    load_size, core_object = core.object_load(data)
    return data[load_size:], core_object


_object_creator_list = []

def object_creator_regist(creator):
    _object_creator_list.append(creator)


def object_loads(data: bytes):
    """バイト列からオブジェクトを再構築する

    Args:
        data(bytes): バイト列
    
    Returns:
        obj (Object): 再構築したオブジェクト
    """
    
    # ヘッダ読み込み
    _, object_name = core.Object.read_header(data)
    split_name = object_name.split('_')
    name = split_name[0]
    dtypes = []
    for dtype_name in split_name[1:]:
        dtypes.append(bb.dtype_from_name(dtype_name))

    # python 内で対応クラスを探索
    for creator in _object_creator_list:
        data, obj = creator(data, name, dtypes)
        if obj is not None:
            return data, obj
    
    # core を探索
    load_size, obj = core.object_load(data)
    if obj is not None:
        return data[load_size:], obj
    
    print('object_loads not object found : %s'%object_name)
    return data, None



def object_load(filename: str):
    """ファイルからオブジェクトを再構築する

    Args:
        filename(str): ファイル名
    
    Returns:
        obj (Object): 再構築したオブジェクト
    """
    with open(filename, 'rb') as f:
        data, obj = bb.object_loads(f.read())
        if data != b'':
            print('[object_loads] warrning: data is too long')
        return obj


# ---- Objectクラス ----

class Object:
    """各クラスの基底クラス
       
       本クラスから派生する各種のクラスにはシリアライズの機能がサポートされる

    """

    def __init__(self, core_object=None):
        self.core_object = core_object

    def set_core(self, core_object):
        self.core_object = core_object

    def get_core(self):
        return self.core_object
    
    def get_object_name(self):
        core_object = self.get_core()
        if core_object is not None:
            return core_object.get_object_name()
        return ''

    def dumps(self) -> bytes:
        """バイトデータにシリアライズ

           モデルのデータをシリアライズして保存するためのバイト配列を生成
        
        Returns:
            data (bytes): Serialize data
        """
        core_model = self.get_core()
        if core_model is not None:
            return core_model.dump_object()
        return b''
        
    def loads(self, data: bytes) -> bytes:
        """バイトデータをロード

           モデルのデータをシリアライズして復帰のバイト配列ロード
        
        Args:
            data (bytes): Serialize data
        """
        core_model = self.get_core()
        if core_model is not None:
            size = core_model.load_object(data)
            return data[size:]
        return data

    def dump(self, filename: str):
        """ファイルに保存

           モデルのデータをシリアライズしてファイルに保存
        
        Args:
            filename (str): ファイル名
        """
        with open(filename, 'wb') as f:
            f.write(self.dumps())


    def load(self, filename: str):
        """ファイルからロード

           ファイルからロード
        
        Args:
            filename (str): ファイル名
        """
        with open(filename, 'rb') as f:
            data = self.loads(f.read())
            if data != b'':
                print('[Model.loads] warrning: data is too long')



""" setup.py for Binary Brain
"""

import sys
import os
from os.path import join as pjoin
import setuptools
from setuptools import setup, Extension
from setuptools import setup, find_packages
from setuptools.command.build_ext import build_ext
import distutils
import subprocess
import urllib.request
import tarfile
import re

# from distutils import ccompiler
# from distutils import unixccompiler
# from distutils import msvccompiler


# change directory
src_path = os.path.dirname(os.path.abspath(sys.argv[0]))
os.chdir(src_path)


# build flags
VERBOSE     = False
WITH_CUDA   = True
WITH_CEREAL = False


# version
major_version = 0
minor_version = 0
revision_number = 0
with open('binarybrain/include/bb/Version.h', 'r', encoding="utf-8") as f:
    for line in f.readlines():
        m = re.match(r'\s*#\s*define\s+BB_MAJOR_VERSION\s+([0-9]+)', line)
        if m:   major_version = int(m.group(1))
        m = re.match(r'\s*#\s*define\s+BB_MINOR_VERSION\s+([0-9]+)', line)
        if m:   minor_version = int(m.group(1))
        m = re.match(r'\s*#\s*define\s+BB_REVISION_NUMBER\s+([0-9]+)', line)
        if m:   revision_number = int(m.group(1))

__version__ = str(major_version) + '.' + str(minor_version) + '.' + str(revision_number)

if VERBOSE:
    print('version = %s' % __version__)




# wget cereal
if WITH_CEREAL:
    with urllib.request.urlopen('https://github.com/USCiLab/cereal/archive/v1.2.2.tar.gz') as r:
        with open('cereal.tar.gz', 'wb') as f:
            f.write(r.read())
    with tarfile.open('./cereal.tar.gz', 'r') as tar:
        tar.extractall('.')

# search CUDA
def find_in_path(name, path):
    for dir in path.split(os.pathsep):
        binpath = pjoin(dir, name)
        if os.path.exists(binpath):
            return os.path.abspath(binpath)
    return None

def search_cuda():
    if sys.platform.startswith('win32') and 'CUDA_PATH' in os.environ:
        cuda_home    = os.environ['CUDA_PATH']
        cuda_bin     = pjoin(cuda_home, 'bin')
        cuda_include = pjoin(cuda_home, 'include')
        cuda_lib     = pjoin(cuda_home, 'lib', 'x64')
        cuda_nvcc    = pjoin(cuda_bin, 'nvcc')
    elif 'CUDAHOME' in os.environ:
        cuda_home = os.environ['CUDAHOME']
        cuda_bin     = pjoin(cuda_home, 'bin')
        cuda_include = pjoin(cuda_home, 'include')
        cuda_lib     = pjoin(cuda_home, 'lib64')
        cuda_nvcc    = pjoin(cuda_bin, 'nvcc')
    else:
        cuda_nvcc = find_in_path('nvcc', os.environ['PATH'])
        if cuda_nvcc is None:
            return None
        cuda_home = os.path.dirname(os.path.dirname(cuda_nvcc))
        cuda_bin     = pjoin(cuda_home, 'bin')
        cuda_include = pjoin(cuda_home, 'include')
        cuda_lib     = pjoin(cuda_home, 'lib64')

    return {'home':cuda_home, 'nvcc':cuda_nvcc, 'include': cuda_include, 'lib': cuda_lib}

if WITH_CUDA:
    CUDA = search_cuda()
else:
    CUDA = None

if CUDA is None:
    print('CUDA is not found.')

class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


# files
sources       = ['binarybrain/src/core_main.cpp']
define_macros = [('BB_ASSERT_EXCEPTION', '1')]
include_dirs  = [distutils.sysconfig.get_python_inc(), get_pybind_include(), get_pybind_include(user=True), 'binarybrain/include']
lib_dirs      = []

if WITH_CEREAL:
    define_macros += [('BB_WITH_CEREAL', '1')]
    include_dirs  += ['cereal-1.2.2/include']

if CUDA is not None:
    sources       += ['binarybrain/src/core_bbcu.cu']
    define_macros += [('BB_WITH_CUDA', '1')]
    include_dirs  += [CUDA['include'], 'binarybrain/cuda']
    lib_dirs      += [CUDA['lib']]

ext_modules = [
    Extension(
        'binarybrain.core',
        sources,
        define_macros=define_macros,
        include_dirs=include_dirs,
        language='c++'
    ),
]


def hook_compiler(self):
    self.src_extensions.append('.cu')
    super_compile_  = self._compile
    super_compile   = self.compile
    super_link      = self.link

    def _compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
        if VERBOSE:
            print('---------------------')
            print('[_compile]')
            print('obj =', obj)
            print('src =', src)
            print('ext =', ext)
            print('cc_args =', cc_args)
            print('extra_postargs =', extra_postargs)
            print('pp_opts =', pp_opts)
            print('---------------------')
        if os.path.splitext(src)[1] == '.cu':
            postargs = extra_postargs['cu']
        elif os.path.splitext(src)[1] == '.cpp':
            postargs = extra_postargs['cc']
        else:
            postargs = []
        super_compile_(obj, src, ext, cc_args, postargs, pp_opts)
    
    def compile(sources,
                output_dir=None, macros=None, include_dirs=None, debug=0,
                extra_preargs=None, extra_postargs=None, depends=None):
        
        if VERBOSE:
            print('---------------------')
            print('[compile]')
            print('sources =', sources)
            print('output_dir =', output_dir)
            print('macros =', macros)
            print('include_dirs =', include_dirs)
            print('debug =', debug)
            print('extra_preargs =', extra_preargs)
            print('extra_postargs =', extra_postargs)
            print('---------------------')

        if self.compiler_type == 'unix':
            return super_compile(sources,
                        output_dir, macros, include_dirs, debug,
                        extra_preargs, extra_postargs, depends)

        if CUDA is not None:
            macros, objects, extra_postargs, _, _ = \
            self._setup_compile(output_dir, macros, include_dirs,
                            sources, depends, extra_postargs)
            
            # macros
            macs = []
            for mac in macros:
                if len(mac) >= 2:
                    macs.append('-D' + mac[0] + '=' + mac[1])
                else:
                    macs.append('-D' + mac[0])

            # includes
            incs = []
            if self.compiler_type == 'msvc':
                incs += ['-I"' + str(inc) + '"' for inc in include_dirs]
            else:
                incs += ['-I' + str(inc) for inc in include_dirs]
            
            # compile
            objects = []
            for src in sources:
                postargs = []
                if os.path.splitext(src)[1] == '.cu':
                    postargs = extra_postargs['cu']
                elif os.path.splitext(src)[1] == '.cpp':
                    postargs = extra_postargs['cc']

                fname, _ = os.path.splitext(os.path.basename(src))
                obj = os.path.join(output_dir, fname + self.obj_extension)
                objects.append(obj)

                args = [CUDA['nvcc'], '-c', '-o', obj] + incs + macs + [src] + postargs
                print(' '.join(args))
                subprocess.call(args)
#               self.spawn(args)

            return objects
        else:
            return super_compile(sources,
                        output_dir, macros, include_dirs, debug,
                        extra_preargs, extra_postargs['cc'], depends)

    def link(target_desc, objects,
             output_filename, output_dir=None, libraries=None,
             library_dirs=None, runtime_library_dirs=None,
             export_symbols=None, debug=0, extra_preargs=None,
             extra_postargs=None, build_temp=None, target_lang=None):
        
        if VERBOSE:
            print('---------------------')
            print('[link]')
            print('target_desc =', target_desc)
            print('objects =', objects)
            print('libraries =', libraries)
            print('library_dirs =', library_dirs)
            print('runtime_library_dirs =', runtime_library_dirs)
            print('export_symbols =', export_symbols)
            print('debug =', debug)
            print('extra_preargs =', extra_preargs)
            print('extra_postargs =', extra_postargs)
            print('build_temp =', build_temp)
            print('target_lang =', target_lang)
            print('---------------------')

        if CUDA is not None:
            libraries, library_dirs, runtime_library_dirs =\
                    self._fix_lib_args(libraries, library_dirs, runtime_library_dirs)
            
#           os.makedirs(os.path.dirname(output_filename), exist_ok=True)
            
            lib_dirs = []
            if self.compiler_type == 'msvc':
                lib_dirs += ['-L"' + str(libdir) + '"' for libdir in library_dirs]
            else:
                lib_dirs += ['-L' + str(libdir) for libdir in library_dirs]
            
            args = [CUDA['nvcc'], '-shared', '-o', output_filename] + objects + lib_dirs + extra_postargs
            print(' '.join(args))
            subprocess.call(args)
#           self.spawn(args)
        else:
            super_link(target_desc, objects,
                output_filename, output_dir, libraries,
                library_dirs, runtime_library_dirs,
                export_symbols, debug, extra_preargs,
                extra_postargs, build_temp, target_lang)

    # hook
    if self.compiler_type == 'unix':
        self._compile = _compile
    self.compile = compile
    self.link = link


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    cc_args = {'unix':[], 'msvc':[]}
    cu_args = {'unix':[], 'msvc':[]}
    ar_args = {'unix':[], 'msvc':[]}
    if CUDA is None:
        # unix(cpu)
        cc_args['unix'] += ['-mavx2', '-mfma', '-fopenmp', '-std=c++14']
        ar_args['unix'] += ['-fopenmp', '-lstdc++', '-lm']
        
        # windows(cpu)
        cc_args['msvc'] += ['/EHsc', '/Oi', '/MT', '/arch:AVX2', '/openmp', '/std:c++14', '/wd"4819"']
        ar_args['msvc'] += []
    else:
        # unix(gpu)
        cc_args['unix'] += ['-gencode=arch=compute_52,code=sm_52',
                            '-gencode=arch=compute_53,code=sm_53',
                            '-gencode=arch=compute_60,code=sm_60',
                            '-gencode=arch=compute_61,code=sm_61',
                            '-gencode=arch=compute_62,code=sm_62',
                            '-gencode=arch=compute_70,code=sm_70',
                            '-gencode=arch=compute_72,code=sm_72',
                            '-gencode=arch=compute_75,code=sm_75',
                            '-Xcompiler', '-pthread',
                            '-Xcompiler', '-mavx2',
                            '-Xcompiler', '-mfma',
                            '-Xcompiler', '-fopenmp',
                            '-Xcompiler', '-std=c++14',
                            '-Xcompiler', '-fPIC' ]
        cu_args['unix'] += ['-gencode=arch=compute_52,code=sm_52',
                            '-gencode=arch=compute_53,code=sm_53',
                            '-gencode=arch=compute_60,code=sm_60',
                            '-gencode=arch=compute_61,code=sm_61',
                            '-gencode=arch=compute_62,code=sm_62',
                            '-gencode=arch=compute_70,code=sm_70',
                            '-gencode=arch=compute_72,code=sm_72',
                            '-gencode=arch=compute_75,code=sm_75',
                            '-std=c++11',
                            '-Xcompiler', '-fPIC' ]
        ar_args['unix'] += ['-Xcompiler', '-pthread',
                            '-Xcompiler', '-fopenmp',
                            '-lstdc++', '-lm', '-lcublas']
        
        # windows(gpu)
        cc_args['msvc'] += ['-O3',
                            '-Xcompiler', '/bigobj',
                            '-Xcompiler', '/EHsc',
                            '-Xcompiler', '/O2',
                            '-Xcompiler', '/Oi',
                            '-Xcompiler', '/FS',
                            '-Xcompiler', '/Zi',
                            '-Xcompiler', '/MT',
                            '-Xcompiler', '/arch:AVX2',
                            '-Xcompiler', '/openmp',
                            '-Xcompiler', '/std:c++14',
                            '-Xcompiler', '/wd\"4819\"']
        cu_args['msvc'] += ['-O3',
                            '-std=c++17',
                            '-gencode=arch=compute_52,code=sm_52',
                            '-gencode=arch=compute_53,code=sm_53',
                            '-gencode=arch=compute_60,code=sm_60',
                            '-gencode=arch=compute_61,code=sm_61',
                            '-gencode=arch=compute_62,code=sm_62',
                            '-gencode=arch=compute_70,code=sm_70',
                            '-gencode=arch=compute_72,code=sm_72',
                            '-gencode=arch=compute_75,code=sm_75',
                            '-Xcompiler', '/bigobj',
                            '-Xcompiler', '/EHsc',
                            '-Xcompiler', '/O2',
                            '-Xcompiler', '/Oi',
                            '-Xcompiler', '/FS',
                            '-Xcompiler', '/Zi',
                            '-Xcompiler', '/MT',
                            '-Xcompiler', '/wd\"4819\"']
        ar_args['msvc'] += ['-lcublas']
    
    if sys.platform == 'darwin':
        darwin_args = ['-stdlib=libc++', '-mmacosx-version-min=10.7']
        cc_args['unix'] += darwin_args
        ar_args['unix'] += darwin_args
    
    
    def build_extensions(self):
        if CUDA is not None:
            self.compiler.set_executable('compiler_so', CUDA['nvcc'])
            self.compiler.set_executable('compiler_cxx', CUDA['nvcc'])
        
        hook_compiler(self.compiler)
        
        ct = self.compiler.compiler_type
        for ext in self.extensions:
            ext.extra_compile_args = {'cc': self.cc_args[ct], 'cu': self.cu_args[ct]}
            ext.extra_link_args = self.ar_args[ct]
        build_ext.build_extensions(self)

package_data = {
    'binarybrain': [
        'include/bb/*.h',
        'include/bbcu/*.h',
        'cuda/*.cu',
        'cuda/*.cuh',
    ],
}

setup(
    name='binarybrain',
    version=__version__,
    author='Ryuji Fuchikami',
    author_email='ryuji.fuchikami@nifty.com',
    url='https://github.com/ryuz/BinaryBrain',
    description='BinaryBrain for Python',
    long_description='',
    ext_modules=ext_modules,
    install_requires=['pybind11>=2.4', 'numpy'],
    setup_requires=['pybind11>=2.4'],
    cmdclass={'build_ext': BuildExt},
    zip_safe=False,
    packages=['binarybrain'],
    package_data=package_data,
)




# -*- coding: utf-8 -*-

import binarybrain      as bb
import binarybrain.core as core
import numpy as np
from typing import List



class LossFunction(bb.Object):
    """LossFunction class
       損失関数の基底クラス
    """
    
    def __init__(self, core_loss=None):
        super(LossFunction, self).__init__(core_object=core_loss)
        
    def clear(self):
        """値のクリア

           集計をクリアする。通常 epoch の単位でクリアして再集計を行う
        """
        self.get_core().clear()
    
    def get(self):
        """値の取得

        Returns:
            loss(float) : 現在までの損失値を返す
        """
        return self.get_core().get_loss()

    def calculate(self, y_buf, t_buf, mini_batch_size=None):
        """損失の計算

            mini_batch_size はメモリなどの都合でミニバッチをさらに分割する場合などに用いる。通常はNoneでよい。

        Args:
            y_buf (FrameBuffer): forward演算結果
            t_buf (FrameBuffer): 教師データ
            mini_batch_size (int): ミニバッチサイズ

        Returns:
            dy_buf (FrameBuffer) : 逆伝搬させる誤差を返す
        """
        if mini_batch_size is None:
            mini_batch_size = t_buf.get_frame_size()
        return bb.FrameBuffer.from_core(self.get_core().calculate_loss(y_buf.get_core(), t_buf.get_core(), mini_batch_size))


class LossMeanSquaredError(LossFunction):
    """LossMeanSquaredError class

        平均二乗誤差(MSE)を計算して誤差として戻す
    """
    
    def __init__(self, dtype=bb.DType.FP32):
        core_loss = bb.search_core_object('LossMeanSquaredError', [dtype]).create()
        super(LossMeanSquaredError, self).__init__(core_loss=core_loss)



class LossCrossEntropy(LossFunction):
    """LossCrossEntropy class

       交差エントロピー誤差を戻す
    """
    
    def __init__(self, dtype=bb.DType.FP32):
        core_loss = bb.search_core_object('LossCrossEntropy', [dtype]).create()
        super(LossCrossEntropy, self).__init__(core_loss=core_loss)


class LossBinaryCrossEntropy(LossFunction):
    """LossBinaryCrossEntropy class
       
       2値の交差エントロピー誤差を戻す
    """
    
    def __init__(self, dtype=bb.DType.FP32):
        core_loss = bb.search_core_object('LossBinaryCrossEntropy', [dtype]).create()
        super(LossBinaryCrossEntropy, self).__init__(core_loss=core_loss)


class LossSoftmaxCrossEntropy(LossFunction):
    """LossSoftmaxCrossEntropy class

       Softmax(活性化) と CrossEntropy(損失関数) の複合
       両者を統合すると計算が簡略化される。
       
       利用に際しては最終段に Softmax が挿入されたのと等価になるので注意すること。
    """
    
    def __init__(self, dtype=bb.DType.FP32):
        core_loss = bb.search_core_object('LossSoftmaxCrossEntropy', [dtype]).create()
        super(LossSoftmaxCrossEntropy, self).__init__(core_loss=core_loss)


class LossSigmoidCrossEntropy(LossFunction):
    """LossSigmoidCrossEntropy class

       Sigmoid(活性化) と BinaryCrossEntropy(損失関数) の複合
       両者を統合すると計算が簡略化される。
       
       利用に際しては最終段に Sigmoid が挿入されたのと等価になるので注意すること。
    """
    
    def __init__(self, dtype=bb.DType.FP32):
        core_loss = bb.search_core_object('LossSigmoidCrossEntropy', [dtype]).create()
        super(LossSigmoidCrossEntropy, self).__init__(core_loss=core_loss)




# coding: utf-8


from binarybrain.system       import *

from binarybrain.dtype        import *

from binarybrain.object       import *

from binarybrain.tensor       import *
from binarybrain.frame_buffer import *
from binarybrain.variables    import *

from binarybrain.models       import *

from binarybrain.losses       import *
from binarybrain.metrics      import *
from binarybrain.optimizer    import *

from binarybrain.storage      import *
from binarybrain.verilog      import *

